#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Cost template definition and examples.
"""

import math
import numpy as np
from numpy import linalg as la
from scipy.special import binom
from functools import partial, partialmethod

from tvopt import solvers, sets, utils



#%% COST TEMPLATES

class Cost():
    r"""
    Template for a cost function.

    This class defines the template for a cost function
    
    .. math:: f : \mathbb{R}^{n_1 \times n_2 \times \ldots} \times \mathbb{R}_+
              \to \mathbb{R} \cup \{ +\infty \}
    
    which depends on the unknown
    :math:`\pmb{x} \in \mathbb{R}^{n_1 \times n_2 \times \ldots}` and,
    optionally, on the time :math:`t \in \mathbb{R}_+`.
    
    `Cost` objects support the following operations:
        
        - negation
        - sum (by another cost or with a scalar),
        - product (by another cost or with a scalar),
        - division and power with a scalar.
    
    A `Cost` object should expose, compatibly with the smoothness degree, the 
    methods `function`, `gradient`, `hessian`, `proximal`. The convention for
    these methods is that the first positional argument is :math:`\pmb{x}`,
    and only a second positional argument is allowed, for :math:`t`. Any other 
    argument should be passed as a keyword argument.
    
    If the cost is time-varying, then it should expose the methods 
    `time_derivative` and `sample`, as well; see methods' documentation for
    the default behavior.
    
    Attributes
    ----------
    dom : sets.Set
        The x domain :math:`\mathbb{R}^{n_1 \times n_2 \times \ldots}`.
    time : sets.T
        The time domain :math:`\mathbb{R}_+`. If the cost is static this is None.
    is_dynamic : bool
        Attribute to check if the cost is static or dynamic.
    smooth : int
        This attribute stores the smoothness degree of the cost, for example
        it is :math:`0` if the cost is continuous, :math:`1` if the cost
        is differentiable, *etc*. By convention it is :math:`-1` if the cost
        is discontinuous.
    prox_solver : str or None
        This attribute specifies the method (gradient or Newton) that should be
        used to compute the proximal
        
        .. math:: \text{prox}_{\rho f(\cdot; t)}(\pmb{x}) = \text{argmin}_{\pmb{y}} 
                  \left\{ f(\pmb{y};t) + \frac{1}{2 \rho} \| \pmb{y} - \pmb{x} \|^2 \right\}
        
        of the cost, if a closed form is not available. See also the auxiliary
        function `compute_proximal`.
    
    Notes
    -----
    Not all operations preserve convexity.
    """
    
    def __init__(self, dom, time=None, prox_solver=None):
        """
        Class constructor.

        Parameters
        ----------
        dom : sets.Set
            The x domain of the cost.
        time : sets.T, optional
            The time domain of the cost. The default is None, in which case the
            cost is static.
        prox_solver : str, optional
            The method for computing the proximal if it is unavailable in 
            closed form. The default is None, and see `compute_proximal` for 
            the behavior in this case.
        """
                
        self.dom, self.time = dom, time
        self.is_dynamic = time is not None
        self.smooth = -1
        self.prox_solver = prox_solver # proximal solver method
    
    def function(self, x, *args, **kwargs):
        """
        An evaluation of the cost. *Implement if needed*.

        Parameters
        ----------
        x : array_like
            The x where the cost should be evaluated.
        *args
            The time at which the cost should be evaluated. Not required if the
            cost is static.
        **kwargs
            Any other required argument.
        """
        
        raise NotImplementedError()
    
    def gradient(self, x, *args, **kwargs):
        """
        An evaluation of the cost's gradient or sub-gradient. *Implement if 
        needed*.

        Parameters
        ----------
        x : array_like
            The x where the (sub-)gradient should be evaluated.
        *args
            The time at which the (sub-)gradient should be evaluated. Not 
            required if the cost is static.
        **kwargs
            Any other required argument.
        """
        
        raise NotImplementedError()
    
    def hessian(self, x, *args, **kwargs):
        """
        An evaluation of the cost's Hessian. *Implement if needed*.

        Parameters
        ----------
        x : array_like
            The x where the Hessian should be evaluated.
        *args
            The time at which the Hessian should be evaluated. Not 
            required if the cost is static.
        **kwargs
            Any other required argument.
        """
        
        raise NotImplementedError()
    
    def proximal(self, x, *args, penalty=1, **kwargs):
        """
        An evaluation of the cost's proximal.
        
        If this method is not overwritten, the default behavior is to 
        recursively compute the proximal via a gradient or Newton backtracking
        algorithm. See `compute_proximal` for the function that is used
        for this purpose.

        Parameters
        ----------
        x : array_like
            The x where the proximal should be evaluated.
        *args
            The time at which the proximal should be evaluated. Not 
            required if the cost is static.
        penalty : float, optional
            The penalty parameter :math:`\rho` for the proximal evaluation.
            Defaults to :math:`1`.
        **kwargs
            Any other required argument.
        """
        
        if self.is_dynamic:
            return self.sample(args[0]).proximal(x, penalty=penalty, **kwargs)
        else:
            if "solver" not in kwargs: kwargs["solver"] = self.prox_solver
            return compute_proximal(self, x, penalty, **kwargs)
    
    def time_derivative(self, x, t, der="tx", **kwargs):
        """
        A derivative w.r.t. time of the cost.
        
        This method computes derivatives w.r.t. time of the cost, or mixed 
        derivatives w.r.t. both time and x (*e.g.* the derivative in time of 
        the gradient).
        
        If this method is not overwritten, it computes the derivative by 
        default using *backward finite differences*. See 
        `backward_finite_difference` for details.
        
        If the cost is static, :math:`0` is returned.

        Parameters
        ----------
        x : array_like
            The x where the derivative should be evaluated.
        t : float
            The time at which the derivative should be evaluated.
        der : str, optional
            A sequence of "x" and "t" that chooses which derivative should be 
            computed. For example, the default "tx" denotes the derivative
            w.r.t. time of the cost's (sub-)gradient.
        **kwargs
            Any other required argument.

        Raises
        ------
        ValueError
            If the number of "x" characters in `der` exceeds :math:`2`.

        Returns
        -------
        array_like
            The required derivative or :math:`0`.
        """
        
        if self.is_dynamic:
            
            # parse the derivative order
            der = "".join(der.split()).lower() # remove spaces, set to lower case
            # num. of derivations w.r.t. time and x
            order_t, order_x = der.count("t"), der.count("x")
            
            # signal to be derived
            if order_x == 0: s = self.function
            elif order_x == 1: s = self.gradient
            elif order_x == 2: s = self.hessian
            else: raise ValueError("Derivative of order {} w.r.t. `x` is unavailable.".format(order_x))
            
            return backward_finite_difference(partial(s, x, **kwargs), t, order=order_t, step=self.time.t_s)
        
        else:
            return 0
            
    def sample(self, t):
        """
        Sample the cost.
        
        This method returns a `SampledCost` object which exposes the same 
        methods of the cost but fixing the time argument to `t`.
        
        If the cost is static, the cost itself is returned.

        Parameters
        ----------
        t : float
            The time at which the cost should be sampled.

        Returns
        -------
        Cost
            The sampled cost or, if static, the cost itself.
        """
        
        if self.is_dynamic: return SampledCost(self, t)
        else: return self
    
    def __add__(self, other):
        
        if utils.is_scalar(other): other = Constant(self.dom, other)
        return SumCost(self, other)
    
    def __mul__(self, other):
        
        if utils.is_scalar(other): return ScaledCost(self, other)
        else: return ProductCost(self, other)
    
    def __pow__(self, other):
        
        return PowerCost(self, other)
    
    def __neg__(self):
        
        return self.__mul__(-1)

    def __radd__(self, other):
        
        return self.__add__(other)
    
    def __sub__(self, other):
        
        return self.__add__(-other)
    
    def __rsub__(self, other):
        
        return self.__neg__().__add__(other)
    
    def __rmul__(self, other):
        
        return self.__mul__(other)
    
    def __truediv__(self, other):
        
        return self.__mul__(1/other)


# -------- SPECIAL COSTS

class SeparableCost(Cost):
    r"""
    Separable cost function.

    This class defines a separable cost, that is
    
    .. math:: f(\pmb{x}; t) = \sum_{i = 1}^N f_i(x_i; t)
    
    where :math:`x_i \in \mathbb{R}^{n_1 \times n_2 \times \ldots}` for each
    :math:`i = 1, \ldots, N`. Each of the component functions :math:`f_i` can
    be either static or dynamic. This is useful for defining distributed
    optimization problems.
    
    The overall dimension of the domain is 
    :math:`n_1 \times n_2 \times \ldots \times N`, meaning that the last 
    dimension indexes the components.
    
    The class exposes the same methods as any `Cost`, with the difference that
    the keyword argument `i` can be used to evaluate only a single component.
    If all components are evaluated, an ndarray is returned with the last
    dimension indexing the components.

    The class has the `Cost` attributes, with the following additions or
    differences.
    
    Attributes
    ----------
    costs : list
        The component costs.
    N : int
        The number of components.
    is_dynamic : bool
        True if at least one component is dynamic.
    smooth : int
        This is the minimum of the smoothness degrees of all components.
    """
    
    def __init__(self, costs):
        """
        Class constructor.

        Parameters
        ----------
        costs : list
            The component costs.
        """
        
        # check if there are dynamic costs
        times = [c.time for c in costs if c.is_dynamic]
        time = times[0] if len(times) > 0 else None
        
        super().__init__(sets.R(*costs[0].dom.shape, len(costs)), time)
        self.costs, self.N = costs, len(costs)
        
        self.smooth = min([c.smooth for c in costs])
    
    def _evaluate(self, method, x, *args, i=None, **kwargs):
        """
        Auxiliary method for components' evaluation.
        
        This method performs an evaluation of the components' function,
        (sub-)gradient or Hessian methods. If the keyword argument `i` is 
        specified, then only the corresponding cost is evaluated.

        Parameters
        ----------
        method : str
            The method to be evaluated.
        x : array_like
            The x where the cost(s) should be evaluated.
        *args
            The time at which the cost(s) should be evaluated. Not required if 
            the cost is static.
        i : int, optional
            If specified, only the corresponding component cost :math:`f_i` is
            evaluated.
        **kwargs
            Any other required argument.

        Returns
        -------
        ndarray
            If `i` is specified, the evaluation of the i-th component, 
            otherwise an ndarray stacking the components evaluations along
            the last dimension.
        """
        
        if i is not None:
            return getattr(self.costs[i], method)(x, *args, **kwargs) if self.costs[i].is_dynamic \
                   else getattr(self.costs[i], method)(x, **kwargs)
        else:
            return np.stack([getattr(c, method)(x[...,i], *args, **kwargs) if c.is_dynamic \
                   else getattr(c, method)(x[...,i], **kwargs) for c, i in zip(self.costs, range(self.N))], axis=-1)  
    
    function = partialmethod(_evaluate, "function")
    gradient = partialmethod(_evaluate, "gradient")
    hessian = partialmethod(_evaluate, "hessian")
    
    def proximal(self, x, *args, penalty=1, i=None, **kwargs):
        """
        An evaluation of the cost(s) proximal(s).
        
        This is the same as calling _evaluate with "proximal", with the 
        difference that is customized to handle the penalty parameter. In 
        particular, the penalty can either be a scalar, in which case the same
        penalty is used for all components, or a list of component-wise 
        penalties.
        """

        if i is None and not isinstance(penalty, list):
            penalty = [penalty for _ in range(self.N)]
        
        # compute the proximals
        if i is not None:
            return self.costs[i].proximal(x, *args, penalty=penalty, **kwargs) if self.costs[i].is_dynamic \
                   else self.costs[i].proximal(x, penalty=penalty, **kwargs)
        else:
            return np.stack([c.proximal(x[...,i], *args, penalty=penalty[i], **kwargs) if c.is_dynamic \
                   else c.proximal(x[...,i], penalty=penalty[i], **kwargs) for c, i in zip(self.costs, range(self.N))], axis=-1) 

class DiscreteDynamicCost(Cost):
    r"""
    Dynamic cost from a sequence of static costs.
    
    This class creates a dynamic cost from a list of static costs. That is, 
    given a sampling time :math:`T_\mathrm{s}`, the cost at time 
    :math:`t_k = k T_\mathrm{s}` is:
    
    .. math:: f(\pmb{x}; t_k) = f_k(\pmb{x})
    
    with :math:`f_k` the k-th static cost in the list.
    """
    
    def __init__(self, costs, t_s=1):
        """
        Class constructor.
        
        Creates the dynamic cost, optionally using the specified sampling
        time `t_s`.

        Parameters
        ----------
        costs : list
            The sequence of static costs.
        t_s : float, optional
            The sampling time, that is, the time that elapses between two
            consecutive costs. The default is 1.
        
        Notes
        -----
        The costs are implicitly assumed to have the same domain and to be 
        static.
        """
        
        # create time domain
        time = sets.T(t_s, t_min=0, t_max=len(costs)*t_s)
        
        super().__init__(costs[0].dom, time)
        self.costs, self.N = costs, len(costs)
        
        self.smooth = min([c.smooth for c in costs])
    
    def _evaluate(self, method, x, t, **kwargs):
        r"""
        Auxiliary method for costs evaluation.
        
        This method evaluates the required method at the given time, by calling
        the method of the k-th function, with :math:`t_k` being the closest
        sampling time to `t`.

        Parameters
        ----------
        method : str
            The method to be evaluated.
        x : array_like
            The x where the cost should be evaluated.
        t : float
            The time at which the cost should be evaluated.
        **kwargs
            Any other required argument.

        Returns
        -------
        ndarray
            The evaluation of the cost.
        """
        
        return getattr(self.costs[self.time.check_input(t)], method)(x, **kwargs)
    
    function = partialmethod(_evaluate, "function")
    gradient = partialmethod(_evaluate, "gradient")
    hessian = partialmethod(_evaluate, "hessian")
    proximal = partialmethod(_evaluate, "proximal")

    def sample(self, t):
        """
        Sample the cost.
        
        The difference with the default `Cost` method is that it returns a
        cost in the list rather than a `SampledCost`.

        Parameters
        ----------
        t : float
            The time at which the cost should be sampled.

        Returns
        -------
        Cost
            The closest cost in the list.
        """
        
        return self.costs[self.time.check_input(t)]


# -------- AUXILIARY CLASSES

class Constant(Cost):
    """
    Constant cost.
    
    This class defines a constant, whose value is stored in the attribute `c`.
    The `gradient` and `hessian` methods return 0, while the proximal acts
    as an identity.
    
    Attributes
    ----------
    dom : sets.Set
        The given cost domain, for compatibility with other costs.
    c : float
        The constant value.
    smooth : int
        The smoothness degree, set to 2.
    """
    
    def __init__(self, dom, c):
        """
        Class constructor.

        Parameters
        ----------
        dom : sets.Set
            The cost domain.
        c : float
            The constant value.
        """
        
        super().__init__(dom)
        self.c = np.array(c).item() # this allows to account for 0-dim or 1 x 1 ndarrays
        
        self.smooth = 2
    
    def function(self, *args, **kwargs):
        """
        An evaluation of the cost.
        
        Returns the costant value.
        """

        return self.c
    
    def gradient(self, *args, **kwargs):
        """
        An evaluation of the cost's gradient.
        
        Returns 0.
        """

        return 0
    
    def hessian(self, *args, **kwargs):
        """
        An evaluation of the cost's Hessian.
        
        Returns 0.
        """
        
        return 0
    
    def proximal(self, x, *args, **kwargs):
        """
        An evaluation of the cost's proximal.
        
        Acts as the identity, returning `x`.
        """
        
        return x

class SumCost(Cost):
    r"""
    Sum of costs.
    
    This class defines a cost as the sum of an arbitrary number of costs. That
    is, given the costs 
    :math:`f_i : \mathbb{R}^n \times \mathbb{R}_+ \to \mathbb{R} \cup \{ +\infty \}`
    with :math:`i = 1, \ldots, N`, the class defines:
        
        .. math:: f(\pmb{x}; t) = \sum_{i = 1}^N f_i(\pmb{x}; t)
    
    The `function`, `gradient` and `hessian` are defined from the components' 
    methods using the sum rule, while the proximal by default is computed 
    recursively.
    """
    
    def __init__(self, *costs):
        """
        Class constructor.
        
        Define the sum of the `costs`, checking if any of them is dynamic.

        Parameters
        ----------
        *costs : Cost
            An arbitrary number of costs to be summed.
        """
        
        # check if dynamic
        times = [c.time for c in costs if c.is_dynamic]
        time = times[0] if len(times) > 0 else None

        super().__init__(costs[0].dom, time)
        self.costs, self.N = costs, len(costs)
        
        self.smooth = min([c.smooth for c in costs])
    
    def _evaluate(self, method, x, *args, **kwargs):
        """
        Auxiliary method for cost evaluation.
        
        An evaluation of the sum of costs `function`, `gradient` or `hessian`.

        Parameters
        ----------
        method : str
            The method to be evaluated.
        x : array_like
            The x where the cost should be evaluated.
        *args
            The time at which the cost(s) should be evaluated. Not required if 
            the cost is static.
        **kwargs
            Any other required argument.

        Returns
        -------
        ndarray
            The evaluation of the sum cost.
        """
        
        return sum([getattr(c, method)(x, *args, **kwargs) if c.is_dynamic \
                    else getattr(c, method)(x, **kwargs) for c in self.costs])
    
    function = partialmethod(_evaluate, "function")
    gradient = partialmethod(_evaluate, "gradient")
    hessian = partialmethod(_evaluate, "hessian")
        
class ScaledCost(Cost):
    r"""
    Scaled cost.
    
    This class defines a cost scaled by a constant. That is, given the cost
    :math:`f : \mathbb{R}^n \times \mathbb{R}_+ \to \mathbb{R} \cup \{ +\infty \}`
    and :math:`c \in \mathbb{R}` it defines:
        
        .. math:: g(\pmb{x}; t) = c f(\pmb{x}; t).
    
    The class is used for the product and division by a constant.
    """
    
    def __new__(cls, cost, s):
        """
        Object instantiation that handles zero scaling constants.
        """
        
        if s == 0: return Constant(cost.dom, 0)
        else: return object.__new__(cls)
        
    def __init__(self, cost, s):
        """
        Class constructor.

        Parameters
        ----------
        cost : Cost
            The cost to be scaled.
        s : float
            The scaling constant.
        """

        # check if dynamic
        time = cost.time if cost.is_dynamic else None
        
        super().__init__(cost.dom, time)
        self.cost, self.s = cost, np.array(s).item()
        
        self.smooth = cost.smooth
    
    def _evaluate(self, method, *args, **kwargs):
        """
        Auxiliary method for cost evaluation.
        
        An evaluation of the scaled cost's `function`, `gradient`, `hessian` or
        `proximal`.

        Parameters
        ----------
        method : str
            The method to be evaluated.
        x : array_like
            The x where the cost should be evaluated.
        *args
            The time at which the cost(s) should be evaluated. Not required if 
            the cost is static.
        **kwargs
            Any other required argument.

        Returns
        -------
        ndarray
            The evaluation of the scaled cost.
        """
        
        # handle the proximal
        if method == "proximal":
            penalty = kwargs.pop("penalty")/self.s
            return getattr(self.cost, method)(*args, penalty=penalty, **kwargs)
        
        return self.s*getattr(self.cost, method)(*args, **kwargs)
    
    function = partialmethod(_evaluate, "function")
    gradient = partialmethod(_evaluate, "gradient")
    hessian = partialmethod(_evaluate, "hessian")
    proximal = partialmethod(_evaluate, "proximal")

class ProductCost(Cost):
    """
    Product of two costs.
    
    This class defines a cost from the product of two given costs. Derivatives
    are computed using the chain rule.
    """
        
    def __init__(self, c_1, c_2):
        """
        Class constructor

        Parameters
        ----------
        c_1 : Cost
            One of the costs to be multiplied.
        c_2 : Cost
            One of the costs to be multiplied.
        """

        # check if dynamic
        times = [c.time for c in (c_1, c_2) if c.is_dynamic]
        time = times[0] if len(times) > 0 else None
        
        super().__init__(c_1.dom, time)
        self.costs = (c_1, c_2)
        
        self.smooth = min(c_1.smooth, c_2.smooth)
    
    def function(self, x, *args, **kwargs):
        """
        An evaluation of the product cost.
        """
        
        f_1, f_2 = [c.function(x, *args, **kwargs) if c.is_dynamic else c.function(x, **kwargs) for c in self.costs]
        
        return f_1*f_2
    
    def gradient(self, x, *args, **kwargs):
        """
        An evaluation of the product cost (sub-)gradient.
        """
        
        f_1, f_2 = [c.function(x, *args, **kwargs) if c.is_dynamic else c.function(x, **kwargs) for c in self.costs]
        g_1, g_2 = [c.gradient(x, *args, **kwargs) if c.is_dynamic else c.gradient(x, **kwargs) for c in self.costs]
        
        return f_2*g_1 + f_1*g_2
    
    def hessian(self, x, *args, **kwargs):
        """
        An evaluation of the product cost Hessian.
        """
        
        f_1, f_2 = [c.function(x, *args, **kwargs) if c.is_dynamic else c.function(x, **kwargs) for c in self.costs]
        g_1, g_2 = [c.gradient(x, *args, **kwargs) if c.is_dynamic else c.gradient(x, **kwargs) for c in self.costs]
        h_1, h_2 = [c.hessian(x, *args, **kwargs) if c.is_dynamic else c.hessian(x, **kwargs) for c in self.costs]
        
        return f_2*h_1 + f_1*h_2 + g_1.dot(g_2.T) + g_2.dot(g_1.T)

class PowerCost(Cost):
    """
    Power cost.
    
    This class defines a cost as the given power of a cost. It is used for 
    implementing the `*` operation.
    """
    # class for costs elevated to some power
    # if p = 0 then a constant cost equal to 1 is returned
    
    def __new__(cls, cost, p):
        """
        Object instantiation that handles zero powers.
        """
        
        if p == 0: return Constant(cost.dom, 1)
        else: return object.__new__(cls)
            
    def __init__(self, cost, p):
        """
        Class constructor.

        Parameters
        ----------
        cost : Cost
            The given cost.
        p : float
            The power for the cost.
        """
        
        p = np.array(p).item()
        
        # check if dynamic
        time = cost.time if cost.is_dynamic else None
        
        super().__init__(cost.dom, time)
        self.cost, self.p = cost, p
        
        self.smooth = cost.smooth
    
    def function(self, *args, **kwargs):
        """
        An evaluation of the power cost.
        """
        
        return self.cost.function(*args, **kwargs)**self.p
    
    def gradient(self, *args, **kwargs):
        """
        An evaluation of the power cost (sub-)gradient.
        """
        
        return self.p*self.cost.function(*args, **kwargs)**(self.p-1) * self.cost.gradient(*args, **kwargs)
    
    def hessian(self, *args, **kwargs):
        """
        An evaluation of the power cost Hessian.
        """
        
        f, g = self.cost.function(*args, **kwargs), self.cost.gradient(*args, **kwargs)
        
        return self.p*(self.p-1)* f**(self.p-2) * np.dot(g, np.transpose(g)) + \
               self.p * f**(self.p-1) * self.cost.hessian(*args, **kwargs)

class SampledCost(Cost):
    """
    Sampled cost.
    
    This class defines a *static* cost by sampling a *dynamic* cost at a given
    time.
    """
                
    def __init__(self, cost, t):
        """
        Class constructor.

        Parameters
        ----------
        cost : Cost
            The dynamic cost to be sampled.
        t : float
            The sampling time.
        """

        super().__init__(cost.dom)
        self.cost, self.t = cost, t
        
        self.smooth = cost.smooth
    
    def _evaluate(self, method, x, **kwargs):
        """
        Auxiliary method for cost evaluation.
        
        An evaluation of the sampled cost's `function`, `gradient`, or 
        `hessian`. Notice that the method does not require positional arguments
        after `x` since the cost is static.

        Parameters
        ----------
        method : str
            The method to be evaluated.
        x : array_like
            The x where the cost should be evaluated.
        **kwargs
            Any other required argument.

        Returns
        -------
        ndarray
            The evaluation of the sampled cost.
        """
        
        return getattr(self.cost, method)(x, self.t, **kwargs)
    
    function = partialmethod(_evaluate, "function")
    gradient = partialmethod(_evaluate, "gradient")
    hessian = partialmethod(_evaluate, "hessian")
    
    def proximal(self, x, penalty=1, **kwargs):
        """
        An evaluation of the cost's proximal.
        """
        
        return super().proximal(x, penalty=penalty, **kwargs)


#%% EXAMPLES: STATIC

# -------- SCALAR COSTS

class AbsoluteValue(Cost):
    """
    Scalar absolute value function.
    """
    
    def __init__(self, weight=1):
        
        super().__init__(sets.R())

        self.weight, self.smooth = weight, 0
    
    def function(self, x):
        
        return self.weight*abs(x)
    
    def gradient(self, x):
        
        return self.weight*math.copysign(1, x)
    
    def proximal(self, x, penalty=1):
        
        return math.copysign(1, x)*max(abs(x) - self.weight*penalty, 0)

class Quadratic_1D(Cost):
    """
    Scalar quadratic cost.
    
    The cost is defined as
    
        .. math:: f(x) = a x^2 / 2 + b x + c.
    """
    
    def __init__(self, a, b, c=0):
        
        super().__init__(sets.R())
        
        self.a, self.b, self.c = np.array(a).item(), np.array(b).item(), np.array(c).item()
        self.smooth = 2
    
    def function(self, x):
        
        return np.array(0.5*self.a*x**2 + self.b*x + self.c).item()
    
    def gradient(self, x):
        
        return self.a*x + self.b
    
    def hessian(self, x=None): # x argument is only for compatibility
        
        return self.a
    
    def proximal(self, x, penalty=1):
        
        return (x - penalty*self.b) / (1 + penalty*self.a)

class Huber_1D(Cost):
    r"""
    Huber loss.
    
    The cost is defined as
    
        .. math:: f(x) = \begin{cases} x^2 / 2 & \text{if} \ |x| \leq \theta \\
                                       \theta (|x| - \theta / 2) & \text{otherwise} \end{cases}
    
    where :math:`\theta > 0` is a given threshold.
    """
    
    def __init__(self, threshold):

        super().__init__(sets.R())
        self.threshold, self.smooth = threshold, 2
    
    def function(self, x):
        
        n = abs(x)
        
        if n <= self.threshold: return x**2/2
        else: return self.threshold*(n - self.threshold/2)
    
    def gradient(self, x):
        
        n = utils.norm(x)
        
        if n <= self.threshold: return x
        else: return self.threshold*x/n
    
    def hessian(self, x):
                
        if abs(x) <= self.threshold: return 1
        else: return 0
    
    def proximal(self, x, penalty=1):
        
        n = abs(x)
        
        if n <= self.threshold*(penalty+1): return x / (penalty+1)
        else: return (1 - penalty*self.threshold/n)*x

class Logistic(Cost):
    r"""
    Logistic function.
    
    The function is defined as
    
        .. math:: f(x) = \log\left( 1 + \exp(x) \right).
    """
    
    def __init__(self):
        
        super().__init__(sets.R())
        self.smooth = 2
    
    def function(self, x):

        return math.log(1 + math.exp(self.dom.check_input(x)))
    
    def gradient(self, x):
        
        x = self.dom.check_input(x)
        
        return math.exp(x) / (1 + math.exp(x))
    
    def hessian(self, x):
        
        x = self.dom.check_input(x)
        
        return math.exp(x) / (1 + math.exp(x))**2


# -------- NORMS

class Norm_1(Cost):
    r"""
    Class for :math:`\ell_1` norm.
    
    The function is defined as
    
        .. math:: f(\pmb{x}) = w \| \pmb{x} \|_1
    
    for :math:`\pmb{x} \in \mathbb{R}^n` and :math:`w > 0`.
    """
    
    def __new__(cls, n=1, weight=1):
        
        if n == 1: return AbsoluteValue(weight=weight)
        else: return object.__new__(cls)
        
    def __init__(self, n=1, weight=1):
        r"""
        Constructor of the cost.
        
        Parameters
        ----------
        n : int
            Size of the unknown :math:`x`.
        weight : float, optional
            A weight multiplying the norm. Defaults to :math:`1`.
        """
        
        # check arguments validity
        super().__init__(sets.R(n, 1))
        
        self.weight = weight
        self.smooth = 0
    
    def function(self, x):
                
        return self.weight*la.norm(self.dom.check_input(x), ord=1)
    
    def gradient(self, x):
                
        return self.weight*np.sign(self.dom.check_input(x))
    
    def proximal(self, x, penalty=1):
        """
        Proximal evaluation of :math:`\ell_1` norm, a.k.a. soft-thresholding.
        
        See Also
        --------
        utils.soft_thresholding
        """
        
        return utils.soft_thresholding(self.dom.check_input(x), self.weight*penalty)

class Norm_2(Cost):
    """
    Square :math:`2`-norm.
    """
    
    def __new__(cls, n=1, weight=1):
        
        return Quadratic(weight*np.eye(n), np.zeros((n, 1)))

class Norm_inf(Cost):
    r"""
    Class for :math:`\ell_\infty` norm.
    """
    
    def __new__(cls, n=1, weight=1):
        
        if n == 1: return AbsoluteValue(weight=weight)
        else: return object.__new__(cls)
    
    def __init__(self, n=1, weight=1):
        
        # check arguments validity
        super().__init__(sets.R(n, 1))
        
        self.weight = weight
        # l1 ball useful for computing the proximal
        self.l1_ball = sets.Ball_l1(np.zeros(self.dom.shape), 1)
    
    def function(self, x):
                
        return self.weight*la.norm(self.dom.check_input(x), ord=np.inf)
    
    def proximal(self, x, penalty=1, tol=1e-5):
        """
        Proximal evaluation of :math:`\ell_\infty` norm.
        
        See [1]__ for the formula.
        
        References
        ----------
        .. [1] A. Beck, First-Order Methods in Optimization. Philadelphia, PA: 
            Society for Industrial and Applied Mathematics, 2017.
        """
        
        # check arguments validity
        x = self.dom.check_input(x)
        
        if np.any(penalty <= 0):
            raise ValueError("The penalty parameter(s) must be positive.")
        
        # compute the proximal
        return x - penalty*self.l1_ball.projection(x / penalty, tol=tol)


# -------- INDICATOR FUNCTIONS

class Indicator(Cost):
    r"""
    Indicator function of a given set.
    
    This objects implements the indicator function of a given `Set` object. 
    That is, given the set :math:`\mathbb{S}` we define:
        
        .. math:: f(\pmb{x}) = \begin{cases} 0 & \text{if} \ \pmb{x} \in \mathbb{S} \\
                                             +\infty & \text{otherwise}. \end{cases}
    
    The proximal operator of the cost is the projection onto the set.
    """
    
    def __init__(self, s):
        
        super().__init__(s)
    
    def function(self, x):
        
        if x in self.dom: return 0
        else: return np.inf
    
    def projection(self, x, **kwargs):
        
        return self.dom.projection(x, **kwargs)
    
    def proximal(self, x, *args, penalty=None, **kwargs):
        
        return self.dom.projection(x, **kwargs)
    
    def __add__(self, other):
        # if other is an indicator, return the indicator of the intersection
        # otherwise normal addition rules apply
        
        if isinstance(other, Indicator): return Indicator(self.dom + other.dom)
        else: return super().__add__(other)


# -------- SMOOTH COSTS

class Linear(Cost):
    r"""
    Linear cost.
    
    The function is defined as
        
            .. math:: f(x) = \langle \pmb{x}, \pmb{b} \rangle + c.
    """
    
    def __new__(cls, a, b=0):
        
        n = np.size(a)
        
        return Quadratic(np.zeros((n, n)), a, b)    

class Quadratic(Cost):
    r"""
    Quadratic cost.
    
    The function is defined as
        
            .. math:: f(x) = \frac{1}{2} \pmb{x}^\top \pmb{A} \pmb{x} + \langle \pmb{x}, \pmb{b} \rangle + c
        
    with the given matrix :math:`\pmb{A} \in \mathbb{R}^{n \times n}` and
    vector :math:`\pmb{b} \in \mathbb{R}^n`.
    """
    
    def __new__(cls, A, b, c=0):
        
        if np.size(A) == 1: return Quadratic_1D(A, b, c)
        else: return object.__new__(cls)
    
    def __init__(self, A, b, c=0):
        
        A = np.array(A)
        super().__init__(sets.R(A.shape[0], 1))
        A = A.reshape((A.shape[0], A.shape[0])) # check it is square
        
        # check b and c (reshaping if need be)
        b = self.dom.check_input(b)
        c = np.array(c).item()
        
        self.A, self.b, self.c = A, b, c
        self.smooth = 2
    
    def function(self, x):
        
        x = self.dom.check_input(x)
        
        return np.array(0.5*x.T.dot(self.A.dot(x)) + self.b.T.dot(x) + self.c).item()
    
    def gradient(self, x):
        
        x = self.dom.check_input(x)
        
        return self.A.dot(x) + self.b
    
    def hessian(self, x=None): # x argument is only for compatibility
        
        return self.A
    
    def proximal(self, x, penalty=1):
        
        x = self.dom.check_input(x)
        
        if self.dom.size == 1: return (x - penalty*self.b) / (1 + penalty*self.A)
        else: return la.solve(np.eye(self.dom.size) + penalty*self.A, x - penalty*self.b)

class Huber(Cost):
    r"""
    Vector Huber loss.
    
    The cost is defined as
    
        .. math:: f(\pmb{x}) = \begin{cases} \|\pmb{x}\|^2 / 2 & \text{if} \ \|\pmb{x}\| \leq \theta \\
                                       \theta (\|\pmb{x}\| - \theta / 2) & \text{otherwise} \end{cases}
    
    where :math:`\theta > 0` is a given threshold.
    """
    
    def __new__(cls, n, threshold):
        
        if n == 1: return Huber_1D(threshold)
        else: return object.__new__(cls)
    
    def __init__(self, n, threshold):
        
        if threshold <= 0: raise ValueError("The threshold must be positive.")
        
        super().__init__(sets.R(n, 1))
        self.threshold = threshold
        self.smooth = 2
    
    def function(self, x):
        
        x = self.dom.check_input(x)
        n = utils.norm(x)
        
        if n <= self.threshold: return n**2/2
        else: return self.threshold*(n - self.threshold/2)
    
    def gradient(self, x):
        
        x = self.dom.check_input(x)
        n = utils.norm(x)
        
        if n <= self.threshold: return x
        else: return self.threshold*x/n
    
    def hessian(self, x):
        
        x = self.dom.check_input(x)
        
        if utils.norm(x) <= self.threshold: return np.eye(self.dom.size)
        else: return np.zeros((self.dom.size, self.dom.size))
    
    def proximal(self, x, penalty=1):
        
        x = self.dom.check_input(x)
        n = utils.norm(x)
        
        if n <= self.threshold*(penalty+1): return x / (penalty+1)
        else: return (1 - penalty*self.threshold/n)*x


# -------- REGRESSION

class LinearRegression(Cost):
    r"""
    Cost for linear regression.
    
    The cost is defined as
    
        .. math:: f(\pmb{x}) = \frac{1}{2} \| \pmb{A} \pmb{x} - \pmb{b} \|^2.
    """
    
    def __new__(cls, A, b):
        
        if np.size(A) == 1: return Quadratic_1D(A**2, -A*b, 0.5*b**2)
        else: return Quadratic(A.T.dot(A), -A.T.dot(b), 0.5*b.T.dot(b))

class RobustLinearRegression(Cost):
    r"""
    Cost for robust linear regression.
    
    Let :math:`h : \mathbb{R} \to \mathbb{R}` be the Huber loss, then thecost 
    is defined as:
    
        .. math:: f(\pmb{x}) = \sum_{i = 1}^m h(a_i \pmb{x} - b_i)
    
    where :math:`a_i \in \mathbb{R}^{1 \times n}` are the rows of the data 
    matrix :math:`\pmb{A} \in \mathbb{R}^{m \times n}`, and :math:`b_i` the 
    elements of the data vector :math:`\pmb{b}`.
    """
    
    def __init__(self, A, b, threshold):
        # each row in A and element of b represent a data point
        
        super().__init__(sets.R(A.shape[1], 1))
        
        self.m = A.shape[0] # domain dim. and num. data points
        # store data as lists
        self.A = [A[d,].reshape(self.dom.shape).T for d in range(self.m)]
        self.b = [np.array(b[d]).item() for d in range(self.m)]
        
        self.huber_fn = Huber(1, threshold) # Huber function
        self.smooth = 2
    
    def function(self, x):
        
        x = self.dom.check_input(x)
        
        return np.sum([self.huber_fn.function(self.A[d].dot(x) - self.b[d]) for d in range(self.m)])
    
    def gradient(self, x):
        
        x = self.dom.check_input(x)
        
        g = np.zeros(self.dom.shape)
        for d in range(self.m):
            
            g += self.huber_fn.gradient(self.A[d].dot(x) - self.b[d])*self.A[d].T

        return g
    
    def hessian(self, x):
        
        x = self.dom.check_input(x)
                
        h = np.zeros(self.dom.shape[:-1] + self.dom.shape[:-1])
        for d in range(self.m):
            
            h += self.huber_fn.hessian(self.A[d].dot(x) - self.b[d])*self.A[d].T.dot(self.A[d])
        
        return h


#%% EXAMPLES: DYNAMIC

class DynamicExample_1D(Cost):
    r"""
    Scalar benchmark dynamic cost.
    
    The dynamic cost was propposed in [1]__ and is defined as:
    
        .. math:: f(x; t) = \frac{1}{2} (x - \cos(\omega t))^2 + \kappa \log(1 + \exp(\mu x))
    
    with default parameters :math:`\omega = 0.02 \pi`, :math:`\kappa = 7.5`
    and :math:`\mu = 1.75`.
    
    .. [1] A. Simonetto, A. Mokhtari, A. Koppel, G. Leus, and A. Ribeiro, 
           "A Class of Prediction-Correction Methods for Time-Varying 
           Convex Optimization," IEEE Transactions on Signal Processing, 
           vol. 64, no. 17, pp. 4576–4591, Sep. 2016.
    """
    
    def __init__(self, t_s, t_max, omega=0.02*math.pi, kappa=7.5, mu=1.75):
        
        super().__init__(sets.R(), sets.T(t_s, t_max=t_max))
        
        self.omega, self.kappa, self.mu = omega, kappa, mu
        self.smooth = 2
    
    def function(self, x, t):
                
        return 0.5*(x - math.cos(self.omega*t))**2 + self.kappa*math.log(1 + math.exp(self.mu*x))
    
    def gradient(self, x, t):
                
        return x - math.cos(self.omega*t) + self.kappa*self.mu*math.exp(self.mu*x) / (1 + math.exp(self.mu*x))

    def hessian(self, x, t=None):
                
        return 1 + self.kappa*(self.mu**2)*math.exp(self.mu*x) / (1 + math.exp(self.mu*x))**2

    def time_derivative(self, x, t, der="tx"):
        
        # parse the derivative order
        der = ''.join(der.split()).lower()
                
        # time der. gradient
        if der == "tx" or der == 'xt': return self.omega*math.sin(self.omega*t)
        else: return super().time_derivative(self.dom.check_input(x), t, der=der)
    
    def approximate_time_derivative(self, x, t, der="tx"):
        
        return super().time_derivative(x, t, der=der)

class DynamicExample_2D(Cost):
    r"""
    Bi-dimensional benchmark dynamic cost.
    
    The dynamic cost was propposed in [1]__ and is defined as:
    
        .. math:: f(\pmb{x}; t) = \frac{1}{2} (x_1 - \exp(\cos(t)))^2 + \frac{1}{2} (x_2 - x_1 \tanh(t))^2
    
    where we used the notation :math:`\pmb{x} = [x_1, x_2]^\top`.
    
    .. [1] Y. Zhang, Z. Qi, B. Qiu, M. Yang, and M. Xiao, "Zeroing Neural 
           Dynamics and Models for Various Time-Varying Problems Solving 
           with ZLSF Models as Minimization-Type and Euler-Type Special 
           Cases [Research Frontier]," IEEE Computational Intelligence 
           Magazine, vol. 14, no. 3, pp. 52–60, Aug. 2019.
    """
    
    def __init__(self, t_s, t_max):
        
        super().__init__(sets.R(2, 1), sets.T(t_s, t_max=t_max))
        self.smooth = 2
    
    def function(self, x, t):
                
        return 0.5*(x[0] - math.exp(math.cos(t)))**2 + 0.5*(x[1] - x[0]*math.tanh(t))**2
    
    def gradient(self, x, t):
               
        th = math.tanh(t)
        
        return np.array([[x[0] - math.exp(math.cos(t)) - th*(x[1] - x[0]*th)],
                         [x[1] - x[0]*th]])

    def hessian(self, x=None, t=None):
        
        th = math.tanh(t)
        
        return np.array([[1 + th**2, -th], [-th, 1]])

    def time_derivative(self, x, t, der="tx"):

        # parse the derivative order
        der = ''.join(der.split()).lower()
                
        # time der. gradient
        if der == "tx" or der == 'xt':
            return np.array([[math.exp(math.cos(t))*math.sin(t) - (x[1] - 2*x[0]*math.tanh(t))/math.cosh(t)**2], 
                             [-x[0]/math.cosh(t)**2]]).reshape((-1,1))
        else:
            return super().time_derivative(x, t, der=der)
    
    def approximate_time_derivative(self, x, t, der="tx"):
        
        return super().time_derivative(x, t, der=der)


#%% UTILITY FUNCTIONS

def compute_proximal(f, x, penalty, solver=None, **kwargs):
    """
    Compute the proximal of a cost.
    
    This function (approximately) computes the proximal of a given cost if 
    there is no closed form solution. The function uses either a Newton method 
    or a gradient method, both with backtracking line search.

    Parameters
    ----------
    f : Cost
        The static cost whose proximal is required.
    x : array_like
        Where the proximal has to be evaluated.
    penalty : float
        The penalty of the proximal.
    solver : str, optional
        The method to use for computing the proximal, Newton or gradient. If 
        not specified, Newton is used for twice differentiable function, 
        gradient otherwise.
    **kwargs : dict
        Parameters for the Newton or gradient method.

    Returns
    -------
    y : ndarray
        The proximal.
    
    See Also
    --------
    solvers.backtracking_gradient
    solvers.newton
    """
    
    # generate cost of the proximal problem
    problem = {"f":f + Quadratic(np.eye(f.dom.size)/penalty, -x/penalty, utils.square_norm(x)/(2*penalty))}
    
    if solver is None and f.smooth >= 2: solver = "newton"
    
    if solver == "newton" or solver == "n":
        y = solvers.newton(problem, **kwargs)
    else:
        y = solvers.backtracking_gradient(problem, **kwargs)
    
    return y

def backward_finite_difference(signal, t, order=1, step=1):
    """
    Compute the backward finite difference of a signal.
    
    This function computes an approximate derivative of a given signal using
    backward finite differences. Given the signal :math:`s(t)`, it computes:
    
        .. math:: s^o(t) = \sum_{i = 0}^o (-1)^i {o \choose i} s(t - i T_s) / T_s^o
    
    where :math:`o \in \mathbb{N}` is the derivative order and :math:`T_s` is 
    the sampling time, see [1]__ for more details.
    
    Notice that if samples before :math:`t = 0` are required, they are set to 
    zero.

    Parameters
    ----------
    signal
        A function of a single scalar argument that represents the signal.
    t : float
        The time where the derivative should be evaluated.
    order : int, optional
        The derivative order, defaults to 1.
    step : float, optional
        The sampling time, defaults to 1.

    Raises
    ------
    ValueError
        For invalid `order` or `step` arguments.

    Returns
    -------
    ndarray
        The approximate derivative.
    
    References
    ----------
    .. [1] A. Quarteroni, R. Sacco, and F. Saleri, Numerical mathematics, 2nd 
           ed. Berlin; New York: Springer, 2007.
    """
        
    if order < 1: raise ValueError("`order` must be a positive integer.")
    if step <= 0: raise ValueError("`step` must be positive.")
    
    terms = []
    for o in range(order+1):
        v = t - o*step # next time
        c = (-1)**o * binom(order, o)
        
        # data point
        if v < 0: terms.append(c*signal(0))
        else: terms.append(c*signal(v))
        
    return sum(terms) / step**order


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Solvers.
"""

import numpy as np
from numpy import linalg as la

from tvopt import utils, costs


#%% FIRST ORDER METHODS

def subgradient(problem, x_0=0, num_iter=100, tol=None):
    r"""
    Sub-gradient method.
    
    This function implements the sub-gradient method
    
        .. math:: \pmb{x}^{\ell+1} = \pmb{x}^\ell - \alpha^\ell \tilde{\nabla} f(\pmb{x}^\ell)
    
    where :math:`\tilde{\nabla} f(\pmb{x}^\ell) \in \partial f(\pmb{x}^\ell)`
    is a sub-differential and :math:`\alpha^\ell = 1 / (\ell + 1)`.

    Parameters
    ----------
    problem : dict
        Problem dictionary defining the cost :math:`f`.
    x_0 : array_like, optional
        The initial condition. This can be either an ndarray of suitable size,
        or a scalar. If it is a scalar then the same initial value is used for
        all components of :math:`\pmb{x}`.
    num_iter : int, optional
        The number of iterations to be performed.
    tol : float, optional
        If given, this argument specifies the tolerance :math:`t` in the 
        stopping condition :math:`\| \pmb{x}^{\ell+1} - \pmb{x}^\ell \| \leq t`.

    Returns
    -------
    x : ndarray
        The approximate solution after `num_iter` iterations.
    """

    f = problem["f"]
    x = np.zeros(f.dom.shape)
    x[...] = x_0
    
    
    for l in range(num_iter):
        
        x_old = x
        
        # choose step-size
        step = 1 / (l+1)
        
        # sub-gradient step
        x = x - step*f.gradient(x)
            
        if stop(x, x_old, tol): break
    
    return x

def gradient(problem, step, x_0=0, num_iter=100, tol=None):
    r"""
    Gradient method.
    
    This function implements the gradient method
    
        .. math:: \pmb{x}^{\ell+1} = \pmb{x}^\ell - \alpha \nabla f(\pmb{x}^\ell)
    
    for a given step-size :math:`\alpha > 0`.

    Parameters
    ----------
    problem : dict
        Problem dictionary defining the cost :math:`f`.
    step : float
        The algorithm's step-size.
    x_0 : array_like, optional
        The initial condition. This can be either an ndarray of suitable size,
        or a scalar. If it is a scalar then the same initial value is used for
        all components of :math:`\pmb{x}`.
    num_iter : int, optional
        The number of iterations to be performed.
    tol : float, optional
        If given, this argument specifies the tolerance :math:`t` in the 
        stopping condition :math:`\| \pmb{x}^{\ell+1} - \pmb{x}^\ell \| \leq t`.

    Returns
    -------
    x : ndarray
        The approximate solution after `num_iter` iterations.
    """
    
    f = problem["f"]
    x = np.zeros(f.dom.shape)
    x[...] = x_0
    
    
    for l in range(num_iter):
        
        x_old = x
        
        # gradient step
        x = x - step*f.gradient(x)
        
        if stop(x, x_old, tol): break
    
    return x

def backtracking_gradient(problem, r=0.2, c=0.5, x_0=0, num_iter=100, tol=None):
    r"""
    Gradient method with backtracking line search.
    
    This function implements the gradient method
    
        .. math:: \pmb{x}^{\ell+1} = \pmb{x}^\ell - \alpha^\ell \nabla f(\pmb{x}^\ell)
    
    where :math:`\alpha^\ell` is chosen via a backtracking line search. In 
    particular, at each iteration we start with :math:`\alpha^\ell = 1` and,
    while
    
        .. math:: f(\pmb{x}^\ell - \alpha^\ell \nabla f(\pmb{x}^\ell)) >
                  f(\pmb{x}^\ell) - c \alpha^\ell \| \nabla f(\pmb{x}^\ell) \|^2
    
    we set :math:`\alpha^\ell = r \alpha^\ell` until a suitable step is found.
    
    Note that the backtracking line search does not stop until a suitable
    step-size  si found; this means that large `r` parameters may result in
    big computation times.

    Parameters
    ----------
    problem : dict
        Problem dictionary defining the cost :math:`f`.
    r : float, optional
        The value by which a candidate step-size is multiplied if it does not
        satisfy the descent condition. `r` should be in :math:`(0,1)`.
    c : float, optional
        The parameter defining the descent condition that a candidate step
        must satisfy.
    x_0 : array_like, optional
        The initial condition. This can be either an ndarray of suitable size,
        or a scalar. If it is a scalar then the same initial value is used for
        all components of :math:`\pmb{x}`.
    num_iter : int, optional
        The number of iterations to be performed.
    tol : float, optional
        If given, this argument specifies the tolerance :math:`t` in the 
        stopping condition :math:`\| \pmb{x}^{\ell+1} - \pmb{x}^\ell \| \leq t`.

    Returns
    -------
    x : ndarray
        The approximate solution after `num_iter` iterations.
    """
        
    f = problem["f"]
    x = np.zeros(f.dom.shape)
    x[...] = x_0
    
    
    for l in range(num_iter):
        
        # candidate descent direction
        grad = f.gradient(x)
        
        # backtracking line search
        step, end = 1, False
        
        x_old, f_old = x, f.function(x)
        while not end:
            
            # candidate point and cost evaluation
            x = x_old - step*grad
            f_new = f.function(x)
            
            if f_new > f_old - c*step*la.norm(grad)**2: step *= r
            else: end = True
            
        if stop(x, x_old, tol): break
    
    return x

def ppa(problem, penalty, x_0=0, num_iter=100, tol=None):
    r"""
    Proximal point algorithm (PPA).
    
    This function implements the proximal point algorithm
    
        .. math:: \pmb{x}^{\ell+1} = \operatorname{prox}_{\rho f}(\pmb{x}^\ell)
    
    where :math:`\rho > 0` is the penalty parameter and we recall that
    
        .. math:: \operatorname{prox}_{\rho f}(\pmb{x}) =
                  \operatorname{arg\,min}_{\pmb{y}} \left\{ f(\pmb{y}) + \frac{1}{2\rho} \| \pmb{y} - \pmb{x} \|^2 \right\}.
    
    Parameters
    ----------
    problem : dict
        Problem dictionary defining the cost :math:`f`.
    penalty : float
        The penalty parameter for the proximal evaluation.
    x_0 : array_like, optional
        The initial condition. This can be either an ndarray of suitable size,
        or a scalar. If it is a scalar then the same initial value is used for
        all components of :math:`\pmb{x}`.
    num_iter : int, optional
        The number of iterations to be performed.
    tol : float, optional
        If given, this argument specifies the tolerance :math:`t` in the 
        stopping condition :math:`\| \pmb{x}^{\ell+1} - \pmb{x}^\ell \| \leq t`.

    Returns
    -------
    x : ndarray
        The approximate solution after `num_iter` iterations.
    """
                
    f = problem["f"]
    x = np.zeros(f.dom.shape)
    x[...] = x_0
    
    
    for l in range(num_iter):
        
        # proximal step
        x_old = x
        x = f.proximal(x, penalty)
        
        if stop(x, x_old, tol): break
    
    return x


# -------- COMPOSITE OPTIMIZATION

def fbs(problem, step, rel=1, x_0=0, num_iter=100, tol=None):
    r"""
    Forward-backward splitting (FBS).
    
    This function implements the forward-backward splitting (a.k.a. proximal 
    gradient method) to solve the composite problem
    
        .. math:: \min_{\pmb{x}} \{ f(\pmb{x}) + g(\pmb{x}) \}.
    
    The algorithm is characterized by the update:
    
        .. math:: \pmb{x}^{\ell+1} = (1-\alpha) \pmb{x}^\ell + 
                  \alpha \operatorname{prox}_{\rho g}(\pmb{x}^\ell - \rho \nabla f(\pmb{x}^\ell))
    
    where :math:`\rho > 0` is the step-size and :math:`\alpha \in (0,1]` is the
    relaxation constant.
    
    Parameters
    ----------
    problem : dict
        Problem dictionary defining the costs :math:`f` and :math:`g`.
    step : float
        The algorithm's step-size.
    rel : float, optional
        The relaxation constant.
    x_0 : array_like, optional
        The initial condition. This can be either an ndarray of suitable size,
        or a scalar. If it is a scalar then the same initial value is used for
        all components of :math:`\pmb{x}`.
    num_iter : int, optional
        The number of iterations to be performed.
    tol : float, optional
        If given, this argument specifies the tolerance :math:`t` in the 
        stopping condition :math:`\| \pmb{x}^{\ell+1} - \pmb{x}^\ell \| \leq t`.

    Returns
    -------
    x : ndarray
        The approximate solution after `num_iter` iterations.
    """
                
    f, g = problem["f"], problem["g"]
    x = np.zeros(f.dom.shape)
    x[...] = x_0
    
    
    for l in range(num_iter):
        
        x_old = x
        
        # gradient and proximal steps
        x = (1-rel)*x + rel*g.proximal(x - step*f.gradient(x), step)
        
        if stop(x, x_old, tol): break
    
    return x

def prs(problem, penalty, rel=1, x_0=0, num_iter=100, tol=None):
    r"""
    Peaceman-Rachford splitting (PRS).
    
    This function implements the Peaceman-Rachford splitting to solve the 
    composite problem
    
        .. math:: \min_{\pmb{x}} \{ f(\pmb{x}) + g(\pmb{x}) \}.
    
    The algorithm is characterized by the updates:
        
        .. math:: \begin{align} \pmb{x}^\ell &= \operatorname{prox}_{\rho f}(\pmb{z}^\ell) \\
                  \pmb{y}^\ell &= \operatorname{prox}_{\rho g}(2 \pmb{x}^\ell - \pmb{z}^\ell) \\
                  \pmb{z}^{\ell+1} &= \pmb{z}^\ell + 2 \alpha (\pmb{y}^\ell - \pmb{x}^\ell)
                  \end{align}
    
    where :math:`\rho > 0` is the penalty and :math:`\alpha \in (0,1]` is the
    relaxation constant.
    
    Parameters
    ----------
    problem : dict
        Problem dictionary defining the costs :math:`f` and :math:`g`.
    penalty : float
        The algorithm's penalty parameter.
    rel : float, optional
        The relaxation constant.
    x_0 : array_like, optional
        The initial condition. This can be either an ndarray of suitable size,
        or a scalar. If it is a scalar then the same initial value is used for
        all components of :math:`\pmb{x}`.
    num_iter : int, optional
        The number of iterations to be performed.
    tol : float, optional
        If given, this argument specifies the tolerance :math:`t` in the 
        stopping condition :math:`\| \pmb{x}^{\ell+1} - \pmb{x}^\ell \| \leq t`.

    Returns
    -------
    x : ndarray
        The approximate solution after `num_iter` iterations.
    """
    
    f, g = problem["f"], problem["g"]

    z = np.zeros(f.dom.shape)
    z[...] = x_0
    
    
    for l in range(num_iter):
        
        z_old = z
        
        # proximal step (f)
        x = f.proximal(z, penalty)
        
        # proximal step (g) and auxiliary update
        z = z + 2*rel*(g.proximal(2*x - z, penalty) - x)
        
        if stop(z, z_old, tol): break
    
    return x, z


#%% SECOND ORDER METHODS

def newton(problem, r=0.2, c=0.5, x_0=0, num_iter=100, tol=None):
    r"""
    Newton method with backtracking line search.
    
    This function implements the Newton method
    
        .. math:: \pmb{x}^{\ell+1} = \pmb{x}^\ell - \alpha^\ell \nabla^2 f(\pmb{x}^\ell)^{-1} \nabla f(\pmb{x}^\ell)
    
    where :math:`\alpha^\ell` is chosen via a backtracking line search. In 
    particular, at each iteration we start with :math:`\alpha^\ell = 1` and,
    while
    
        .. math:: f(\pmb{x}^\ell - \alpha^\ell \nabla^2 f(\pmb{x}^\ell)^{-1} \nabla f(\pmb{x}^\ell)) >
                  f(\pmb{x}^\ell) - c \alpha^\ell \| \nabla f(\pmb{x}^\ell) \|^2
    
    we set :math:`\alpha^\ell = r \alpha^\ell` until a suitable step is found.
    
    Note that the backtracking line search does not stop until a suitable
    step-size  si found; this means that large `r` parameters may result in
    big computation times.

    Parameters
    ----------
    problem : dict
        Problem dictionary defining the cost :math:`f`.
    r : float, optional
        The value by which a candidate step-size is multiplied if it does not
        satisfy the descent condition. `r` should be in :math:`(0,1)`.
    c : float, optional
        The parameter defining the descent condition that a candidate step
        must satisfy.
    x_0 : array_like, optional
        The initial condition. This can be either an ndarray of suitable size,
        or a scalar. If it is a scalar then the same initial value is used for
        all components of :math:`\pmb{x}`.
    num_iter : int, optional
        The number of iterations to be performed.
    tol : float, optional
        If given, this argument specifies the tolerance :math:`t` in the 
        stopping condition :math:`\| \pmb{x}^{\ell+1} - \pmb{x}^\ell \| \leq t`.

    Returns
    -------
    x : ndarray
        The approximate solution after `num_iter` iterations.
    """
                
    f = problem["f"]
    x = np.zeros(f.dom.shape)
    x[...] = x_0
    
    
    for l in range(num_iter):
        
        # candidate descent direction
        grad, hess = f.gradient(x), f.hessian(x)
        d = utils.solve(hess, grad)
        
        # backtracking line search
        step, end = 1, False
        
        x_old, f_old = x, f.function(x)
        while not end:
            
            # candidate point and cost evaluation
            x = x_old - step*d
            f_new = f.function(x)
            
            if f_new > f_old - c*step*np.dot(np.transpose(grad), d): step *= r
            else: end = True
        
        if stop(x, x_old, tol): break
    
    return x


#%% DUAL-BASED METHODS

def dual_ascent(problem, penalty, w_0=0, num_iter=100, tol=None):
    r"""
    Dual ascent.
    
    This function implements the dual ascent to solve the constrained problem
    
        .. math:: \min_{\pmb{x}} f(\pmb{x}) \ \text{s.t.} \ \pmb{A} \pmb{x} = \pmb{c}.
    
    The algorithm is characterized by the updates:    
    
        .. math:: \begin{align}
                  \pmb{x}^\ell &= \operatorname{arg\,min}_{\pmb{x}} \left\{ f(\pmb{x}) - \langle \pmb{w}^\ell, \pmb{A} \pmb{x} \rangle \right\} \\
                  \pmb{w}^{\ell+1} &= \pmb{w}^\ell - \rho (\pmb{A} \pmb{x}^\ell - \pmb{c})
                  \end{align}
        
    for a given penalty :math:`\rho > 0`.

    Parameters
    ----------
    problem : dict
        Problem dictionary defining the cost :math:`f`, and the constraints :math:`A` and :math:`c`.
    penalty : float
        The algorithm's penalty.
    w_0 : array_like, optional
        The dual initial condition. This can be either an ndarray of suitable 
        size, or a scalar. If it is a scalar then the same initial value is 
        used for all components of :math:`\pmb{w}`.
    num_iter : int, optional
        The number of iterations to be performed.
    tol : float, optional
        If given, this argument specifies the tolerance :math:`t` in the 
        dual stopping condition 
        :math:`\| \pmb{w}^{\ell+1} - \pmb{w}^\ell \| \leq t`.

    Returns
    -------
    x : ndarray
        The approximate primal solution after `num_iter` iterations.
    w : ndarray
        The approximate dual solution after `num_iter` iterations.
    """
    
    f, A, c = problem["f"], problem["A"], problem["c"]
    
    x = np.zeros(f.dom.shape)
    w = np.zeros((A.shape[0],1)) # dual variables
    w[...] = w_0
    
    # auxiliary function for primal update
    p = costs.Linear(- A.T.dot(w))
    
    
    for l in range(num_iter):
               
        # primal update
        x = newton({"f":f+p}, x_0=x, tol=1e-5)
        
        w_old = w
        
        # dual update
        w = w - penalty*(A.dot(x) - c)
        
        if stop(w, w_old, tol): break
        
        # update auxliary function
        p.a = - A.T.dot(w)
    
    return x, w

def mm(problem, penalty, w_0=0, num_iter=100, tol=None):
    r"""
    Method of multipliers (MM).
    
    This function implements the method of multipliers to solve the constrained
    problem
    
        .. math:: \min_{\pmb{x}} f(\pmb{x}) \ \text{s.t.} \ \pmb{A} \pmb{x} = \pmb{c}.
    
    The algorithm is characterized by the updates:    
    
        .. math:: \begin{align}
                  \pmb{x}^\ell &= \operatorname{arg\,min}_{\pmb{x}} \left\{ f(\pmb{x}) 
                  - \langle \pmb{w}^\ell, \pmb{A} \pmb{x} \rangle + \frac{\rho}{2} \| \pmb{A} \pmb{x} - \pmb{c} \|^2 \right\} \\
                  \pmb{w}^{\ell+1} &= \pmb{w}^\ell - \rho (\pmb{A} \pmb{x}^\ell - \pmb{c})
                  \end{align}
        
    for a given penalty :math:`\rho > 0`.

    Parameters
    ----------
    problem : dict
        Problem dictionary defining the cost :math:`f`, and the constraints :math:`A` and :math:`c`.
    penalty : float
        The algorithm's penalty.
    w_0 : array_like, optional
        The dual initial condition. This can be either an ndarray of suitable 
        size, or a scalar. If it is a scalar then the same initial value is 
        used for all components of :math:`\pmb{w}`.
    num_iter : int, optional
        The number of iterations to be performed.
    tol : float, optional
        If given, this argument specifies the tolerance :math:`t` in the 
        dual stopping condition 
        :math:`\| \pmb{w}^{\ell+1} - \pmb{w}^\ell \| \leq t`.

    Returns
    -------
    x : ndarray
        The approximate primal solution after `num_iter` iterations.
    w : ndarray
        The approximate dual solution after `num_iter` iterations.
    """
    
    f, A, c = problem["f"], problem["A"], problem["c"]
    
    x = np.zeros(f.dom.shape)
    w = np.zeros((A.shape[0],1)) # dual variables
    w[...] = w_0
    
    # auxiliary function for primal update
    p = costs.Quadratic(penalty*A.T.dot(A), -penalty*A.T.dot(c + w/penalty))
    
    
    for l in range(num_iter):
                
        # primal update
        x = newton({"f":f+p}, x_0=x, tol=1e-5)
        
        w_old = w
                
        # dual update
        w = w - penalty*(A.dot(x) - c)
        
        if stop(w, w_old, tol): break
        
        # update auxliary function
        p.b = -penalty*A.T.dot(c + w/penalty)
    
    return x, w

def dual_fbs(problem, penalty, rel=1, w_0=0, num_iter=100, tol=None):
    r"""
    Dual forward-backward splitting.
    
    This function implements the dual FBS to solve the constrained problem
    
        .. math:: \begin{align}
                  &\min_{\pmb{x},\pmb{y}} \left\{ f(\pmb{x}) + g(\pmb{y}) \right\} \\
                  &\text{s.t.} \ \pmb{A} \pmb{x} +  \pmb{B} \pmb{y} = \pmb{c}.
                  \end{align}
    
    The algorithm is characterized by the updates:    
    
        .. math:: \begin{align}
                  \pmb{x}^\ell &= \operatorname{arg\,min}_{\pmb{x}} \left\{ f(\pmb{x}) - \langle \pmb{w}, \pmb{A} \pmb{x} \rangle \right\} \\
                  \pmb{u}^\ell &= \pmb{w}^\ell - \rho (\pmb{A} \pmb{x}^\ell - \pmb{c}) \\
                  \pmb{y}^\ell &= \operatorname{arg\,min}_{\pmb{y}} \left\{ g(\pmb{y}) 
                  - \langle \pmb{u}^\ell, \pmb{B} \pmb{y} \rangle + \frac{\rho}{2} \| \pmb{B} \pmb{y} \|^2 \right\} \\
                  \pmb{w}^{\ell+1} &= (1-\alpha) \pmb{w}^\ell + \alpha (\pmb{u}^\ell - \rho \pmb{B} \pmb{y}^\ell)
                  \end{align}
        
    for a given penalty :math:`\rho > 0` and :math:`\alpha \in (0,1]` is the
    relaxation constant.

    Parameters
    ----------
    problem : dict
        Problem dictionary defining the costs :math:`f` and :math:`g`, and the constraints
        :math:`A`, :math:`B` and :math:`c`.
    penalty : float
        The algorithm's penalty.
    rel : float, optional
        The relaxation constant.
    w_0 : array_like, optional
        The dual initial condition. This can be either an ndarray of suitable 
        size, or a scalar. If it is a scalar then the same initial value is 
        used for all components of :math:`\pmb{w}`.
    num_iter : int, optional
        The number of iterations to be performed.
    tol : float, optional
        If given, this argument specifies the tolerance :math:`t` in the 
        dual stopping condition 
        :math:`\| \pmb{w}^{\ell+1} - \pmb{w}^\ell \| \leq t`.

    Returns
    -------
    x : ndarray
        The approximate primal solution :math:`\pmb{x}` after `num_iter` iterations.
    y : ndarray
        The approximate primal solution :math:`\pmb{y}` after `num_iter` iterations.
    w : ndarray
        The approximate dual solution after `num_iter` iterations.
    """
    
    f, g, A, B, c = problem["f"], problem["g"], problem["A"], problem["B"], problem["c"]
    
    x, y = np.zeros(f.dom.shape), np.zeros(g.dom.shape)
    w = np.zeros((A.shape[0],1)) # dual variables
    w[...] = w_0
    
    # auxiliary functions for primal updates
    p_x = costs.Linear(- A.T.dot(w))
    p_y = costs.Quadratic(penalty*B.T.dot(B), np.zeros(g.dom.shape))
    
    
    for l in range(num_iter):
                
        # primal update (x)
        x = newton({"f":f+p_x}, x_0=x, tol=1e-5)
        
        # auxiliary update
        u = w - penalty*(A.dot(x) - c)
        
        # update auxiliary function (y)
        p_y.b = -penalty*B.T.dot(u/penalty)
        
        # primal update (y)
        y = newton({"f":g+p_y}, x_0=y, tol=1e-5)
        
        w_old = w        
        
        # dual update
        w = (1-rel)*w + rel*(u - penalty*B.dot(y))
        
        if stop(w, w_old, tol): break
        
        # update auxliary function (x)
        p_x.a = - A.T.dot(w)
    
    return x, y, w

def admm(problem, penalty, rel=1, w_0=0, num_iter=100, tol=None):
    r"""
    Alternating direction method of multipliers (ADMM).
    
    This function implements the ADMM to solve the constrained problem
    
        .. math:: \begin{align}
                  &\min_{\pmb{x},\pmb{y}} \left\{ f(\pmb{x}) + g(\pmb{y}) \right\} \\
                  &\text{s.t.} \ \pmb{A} \pmb{x} +  \pmb{B} \pmb{y} = \pmb{c}.
                  \end{align}
    
    The algorithm is characterized by the updates:    
    
        .. math:: \begin{align}
                  \pmb{x}^\ell &= \operatorname{arg\,min}_{\pmb{x}} \left\{ f(\pmb{x}) 
                  - \langle \pmb{z}^\ell, \pmb{A} \pmb{x} \rangle + \frac{\rho}{2} \| \pmb{A} \pmb{x} - \pmb{c} \|^2 \right\} \\
                  \pmb{w}^\ell &= \pmb{z}^\ell - \rho (\pmb{A} \pmb{x}^\ell - \pmb{c}) \\
                  \pmb{y}^\ell &= \operatorname{arg\,min}_{\pmb{y}} \left\{ g(\pmb{y}) 
                  - \langle 2 \pmb{w}^\ell - \pmb{z}^\ell, \pmb{B} \pmb{y} \rangle + \frac{\rho}{2} \| \pmb{B} \pmb{y} \|^2 \right\} \\
                  \pmb{u}^\ell &= 2 \pmb{w}^\ell - \pmb{z}^\ell - \rho \pmb{B} \pmb{y}^\ell \\
                  \pmb{z}^{\ell+1} &= \pmb{z}^\ell + 2 \alpha (\pmb{u}^\ell - \pmb{w}^\ell)
                  \end{align}
        
    for a given penalty :math:`\rho > 0` and :math:`\alpha \in (0,1]` is the
    relaxation constant.

    Parameters
    ----------
    problem : dict
        Problem dictionary defining the costs :math:`f` and :math:`g`, and the constraints
        :math:`A`, :math:`B` and :math:`c`.
    penalty : float
        The algorithm's penalty.
    rel : float, optional
        The relaxation constant.
    w_0 : array_like, optional
        The dual initial condition. This can be either an ndarray of suitable 
        size, or a scalar. If it is a scalar then the same initial value is 
        used for all components of :math:`\pmb{w}`.
    num_iter : int, optional
        The number of iterations to be performed.
    tol : float, optional
        If given, this argument specifies the tolerance :math:`t` in the 
        dual stopping condition 
        :math:`\| \pmb{w}^{\ell+1} - \pmb{w}^\ell \| \leq t`.

    Returns
    -------
    x : ndarray
        The approximate primal solution :math:`\pmb{x}` after `num_iter` iterations.
    y : ndarray
        The approximate primal solution :math:`\pmb{y}` after `num_iter` iterations.
    w : ndarray
        The approximate dual solution after `num_iter` iterations.
    """
    
    f, g, A, B, c = problem["f"], problem["g"], problem["A"], problem["B"], problem["c"]
    
    x, y = np.zeros(f.dom.shape), np.zeros(g.dom.shape)
    z = np.zeros((A.shape[0],1)) # auxiliary variables
    z[...] = w_0
    
    # auxiliary functions for primal updates
    p_x = costs.Quadratic(penalty*A.T.dot(A), -penalty*A.T.dot(c + z/penalty))
    p_y = costs.Quadratic(penalty*B.T.dot(B), np.zeros(g.dom.shape))
    
    
    for l in range(num_iter):
                
        # primal update (x)
        x = newton({"f":f+p_x}, x_0=x, tol=1e-5)
        
        # dual update
        w = z - penalty*(A.dot(x) - c)
        
        # update auxiliary function (y)
        p_y.b = -penalty*B.T.dot((2*w - z)/penalty)
        
        # primal update (y)
        y = newton({"f":g+p_y}, x_0=y, tol=1e-5)
    
        z_old = z
        
        # auxiliary update
        z = (1-2*rel)*z + 2*rel*(w - penalty*B.dot(y))
        
        if stop(z, z_old, tol): break
        
        # update auxliary function (x)
        p_x.b = -penalty*A.T.dot(c + z/penalty)
    
    return x, y, z


#%% UTILITY FUNCTIONS

def stop(x, x_old, tol=None):
    """
    Stopping condition.
    
    This function checks the stopping condition
    
        .. math:: \| \pmb{x}^{\ell+1} - \pmb{x}^\ell \| \leq t
    
    if `t` is specified.

    Parameters
    ----------
    x : ndarray
        The current iterate.
    x_old : ndarray
        The previous iterate.
    tol : float, optional
        The tolerance in the stopping condition.

    Returns
    -------
    bool
        True if `tol` is given and the stopping condition is verified, False
        otherwise.
    """

    return tol is not None and la.norm(x - x_old) <= tol


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Utility tools.
"""


import sys
import math
import numbers
import numpy as np
from numpy import linalg as la
from numpy.random import default_rng
from scipy.linalg import block_diag

ran = default_rng() # random generator


#%% LINEAR ALGEBRA

def square_norm(x):
    """
    Compute the square norm of the given vector.

    Parameters
    ----------
    x : array_like
        The vector array.

    Returns
    -------
    ndarray
        The square norm.
    
    Notes
    -----
    The function reshapes `x` to a column vector, so it does not correctly
    handle n-dimensional arrays. For n-dim arrays use `numpy.linalg.norm`.
    """
    
    x = np.reshape(x, (-1,1))
    
    return x.T.dot(x).item()

def norm(x):
    """
    Compute the norm of the given vector.

    Parameters
    ----------
    x : array_like
        The vector array.

    Returns
    -------
    ndarray
        The square norm.
    
    See Also
    --------
    square_norm : Square norm
    
    Notes
    -----
    The function reshapes `x` to a column vector, so it does not correctly
    handle n-dimensional arrays. For n-dim arrays use `numpy.linalg.norm`.
    """
    
    return math.sqrt(square_norm(x))

def normalize(x):
    """
    Normalize a vector to unit vector.

    Parameters
    ----------
    x : array_like
        The vector array.

    Returns
    -------
    ndarray
        The normalized vector.
    
    Notes
    -----
    The function reshapes `x` to a column vector, so it does not correctly
    handle n-dimensional arrays. For n-dim arrays use `numpy.linalg.norm`.
    """
    
    return x / norm(x)

def is_square(mat):
    """
    Check if the matrix is 2-D and square.
    
    Parameters
    ----------
    mat : ndarray
        The given matrix.

    Returns
    -------
    bool
        True if the matrix is 2-D and square, False otherwise.
    """
    
    return len(mat.shape) == 2 and mat.shape[0] == mat.shape[1]

def is_stochastic(mat, row=True, col=True):
    """
    Verify if a given matrix is row, column or doubly stochastic.
    
    Parameters
    ----------
    mat : ndarray
        The given matrix.
    row : bool, optional
        Check for row stochasticity, default True.
    col : bool, optional
        Check for column stochasticity, default True.

    Returns
    -------
    bool
        True if the matrix is stochastic (row, column or doubly, as specified
        by the arguments).
    
    Raises
    ------
    ValueError
        If neither `row` nor `col` are True.
    """
    
    if not row and not col:
        raise ValueError("At least one of row and col must be True.")
    
    # verify the matrix is non-negative
    if np.any(mat < 0):
        return False
    
    # verify row stochasticity
    ones = np.ones((mat.shape[0],1))
    is_row = np.allclose(mat.dot(ones) - ones, np.zeros((mat.shape[0],1)))
    
    # verify column stochasticity
    ones = np.ones((mat.shape[1],1))
    is_col = np.allclose(mat.T.dot(ones) - ones, np.zeros((mat.shape[1],1)))
    
    if row and col: return is_row and is_col
    elif row and not col: return is_row
    else: return is_col

def orthonormal_matrix(dim):
    """
    Generate a random orthonormal matrix.
    
    This function generates uniformly distributed random orthonormal matrices
    using Householder reflections (see Section 7 of `this paper
    <https://arxiv.org/pdf/math-ph/0609050.pdf>`_).
    
    Parameters
    ----------
    dim : int
        Size of the matrix.

    Returns
    -------
    orth_mat : ndarray
        The random orthonormal matrix.
    
    Raises
    ------
    ValueError
        For invalid `dim`.
    """
    
    # check validity of dim and cast to integer
    if int(dim) < 1:
        raise ValueError("The dimension must be at least one.")
    dim = int(dim)
    
    # array for the orthonormal matrix
    orth_mat = np.eye(dim)
        
    for i in range(0,dim):
        
        rnd_vec = ran.standard_normal((dim-i,1)) # generate a random vector
        # first vector in standard basis of R^{dim-i}
        strd_1 = np.vstack((1,np.zeros((dim-i-1,1))))
        # temporary vector for computing HH reflection
        tmp_vec = (rnd_vec + np.sign(rnd_vec[0])*strd_1) \
                / la.norm(rnd_vec + np.sign(rnd_vec[0])*strd_1)
        
        # Householder reflection
        hh_refl = -np.sign(rnd_vec[0])*(np.eye(dim-i)-2*tmp_vec.dot(tmp_vec.T))
        # block diagonal with identity and Householder reflection
        hh_tilde = block_diag(np.eye(i),hh_refl)
        
        # update the orthogonal matrix
        orth_mat = orth_mat.dot(hh_tilde.T)
    
    return orth_mat

def random_matrix(eigs):
    """
    Generate a random matrix.
    
    The matrix is generated as
    
        .. math:: M = O \mathrm{diag}\{ \lambda_i \} O^\\top
    
    where :math:`O` is a random orthonormal matrix and :math:`\lambda_i` are
    the specified eigenvalues.
    
    Parameters
    ----------
    eigs : array-like
        The list of eigenvalues for the matrix.

    Returns
    -------
    ndarray
        The random positive semi-definite matrix.
    
    See Also
    --------
    orthonormal_matrix : Orthonormal matrix generator.
    """
    
    eigs = np.ravel(eigs) # flatten eigs in an array

    # generate an orthonormal matrix
    orth_mat = orthonormal_matrix(eigs.size)
    
    return orth_mat.dot(np.diag(eigs).dot(orth_mat.T))

def positive_semidefinite_matrix(dim, max_eig=None, min_eig=None):
    """
    Generate a random positive semi-definite matrix.
    
    The matrix is generated as
    
        .. math:: M = O \mathrm{diag}\{ \lambda_i \} O^\\top
    
    where :math:`O` is a random orthonormal matrix and :math:`\lambda_i` are
    random eigenvalues uniformly drawn between `min_eig` and `max_eig`. If 
    `dim` is larger than or equal to two, `min_eig` and `max_eig` are included 
    in the eigenvalues list.
    
    Parameters
    ----------
    dim : int
        Size of the matrix.
    eigs : array-like, optional
        The list of eigenvalues for the matrix; if None, the eigenvalues are
        uniformly drawn from :math:`[10^{-2}, 10^2]`.

    Returns
    -------
    ndarray
        The random positive semi-definite matrix.
    
    Raises
    ------
    ValueError.
    
    See Also
    --------
    random_matrix : Random matrix generator.
    """
    
    max_eig = 1e2 if max_eig is None else max_eig
    min_eig = 1e-2 if min_eig is None else min_eig
    
    if max_eig < 0 or min_eig < 0:
        raise ValueError("`max_eig` and `min_eig` must be non-negative.")
    
    if dim < 2:
        eigs = [max_eig]
    if dim < 3:
        eigs = [max_eig, min_eig]
    else:
        eigs = np.hstack((max_eig, (max_eig-min_eig)*ran.random(dim-2)+min_eig, min_eig))
    
    return random_matrix(eigs)

def solve(a, b):
    
    if is_scalar(a) or np.size(a) == 1: return b / a
    else: return la.solve(a, b)
    

#%% PERFORMANCE METRICS

def fpr(s, ord=2):
    """
    Fixed point residual.
    
    This function computes the fixed point residual of a signal `s`,
    that is
    
        .. math:: \{ \| s^\ell - s^{\ell-1} \|_i \}_{\ell \in \mathbb{N}}.
    
    Different norm orders can be used, that can be specified using the
    `numpy.linalg.norm` argument `ord`.

    Parameters
    ----------
    s : array_like
        The signal, with the last dimension indexing time.
    ord : optional
        Norm order, see `numpy.linalg.norm`.

    Returns
    -------
    ndarray
        The fixed point residual.
    """
    
    if isinstance(s, list): s = np.stack(s, axis=-1)
    
    return dist(s[...,:-1], s[...,1:], ord=ord)


def dist(s, r, ord=2):
    """
    Distance of a signal from a reference.
    
    This function computes the distance of a signal `s` from a reference `r`.
    The reference can be either constant or a signal itself.
    Different norm orders can be used, that can be specified using the
    `numpy.linalg.norm` argument `ord`.

    Parameters
    ----------
    s : array_like
        The signal, with the last dimension indexing time.
    r : array_like
        The reference, either a single array or a signal with the last 
        dimension indexing time.
    ord : optional
        Norm order, see `numpy.linalg.norm`.

    Raises
    ------
    ValueError
        For incompatible dimensions of signal and reference.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    
    if isinstance(s, list): s = np.stack(s, axis=-1)
    if isinstance(r, list): r = np.stack(r, axis=-1)
    
    # repeat x_opt if it is a static quantity
    if r.shape == s.shape[:-1]:
        r = np.repeat(r[...,np.newaxis], s.shape[-1], axis=-1)
    if s.ndim == 2 and r.shape[0] == s.shape[0] and r.shape[0] != s.shape[0]:
        r = np.repeat(r, s.shape[-1], axis=-1)
    
    # compute the norm if the shapes are compatible
    if r.shape == s.shape:
        # handle scalar signals
        if s.shape[:-1] == ():
            return np.abs(s - r)
        else:
            d = np.zeros(s.shape[-1])
            for l in range(s.shape[-1]):
                d[l] = la.norm((s[...,l] - r[...,l]).flatten(), ord=ord)
            return d
    else:
        raise ValueError("Incompatible shapes of `s` and `r` {}, {}.".format(s.shape, r.shape))

def regret(f, s, r=None):
    r"""
    Cost over time or regret.
    
    This function computes the cost evaluated using `f` incurred by an 
    approximate minimizer `s`
    
        .. math:: \{ \frac{1}{\ell} \sum_{j = 1}^\ell f(s^j) \}_{\ell \in \mathbb{N}}
    
    or, if a reference `r` is specified, then the function computes the regret
    
        .. math:: \{ \frac{1}{\ell} \sum_{j = 1}^\ell f(s^j) - f(r^j) \}_{\ell \in \mathbb{N}}
    
    where `r` is either a constant array or a signal.    

    Parameters
    ----------
    f : costs.Cost
        The cost to evaluate in the signal.
    s : array_like
        The sequence of approximate minimizers.
    r : array_like, optional
        The reference, either a single array or a signal with the last 
        dimension indexing time.

    Returns
    -------
    ndarray
        The sequence of cost evaluations or regret.
    """
    
    if isinstance(s, list): s = np.stack(s, axis=-1)
    if isinstance(r, list): r = np.stack(r, axis=-1)

        
    reg = np.zeros(s.shape[-1])
    for k in range(s.shape[-1]):
        
        # dynamic cost
        if f.is_dynamic:
            t = k*f.time.t_s
            reg[k] = f.function(s[...,k], t)
            if r is not None: reg[k] -= f.function(r[...,k], t)
        # static cost
        else:
            reg[k] = f.function(s[...,k])
            if r is not None: reg[k] -= f.function(r)
    
    # compute cumulative regret
    cum_reg = np.cumsum(reg)
    return np.array([cum_reg[k]/(k+1) for k in range(s.shape[-1])])


#%% VARIOUS  

def uniform_quantizer(x, step, thresholds=None):
    """
    Function to perform uniform quantization.
    
    The function applies the uniform quantization
    
    .. math:: q(x) = \Delta \operatorname{floor}
              \left( \\frac{x}{\Delta} + \\frac{1}{2} \\right)
    
    where :math:`\Delta` is the given step. Moreover, a saturation to
    upper and lower thresholds is peformed if given as argument.

    Parameters
    ----------
    x : ndarray
        The array to be quantized.
    step : float
        The step of the quantizer.
    thresholds : list, optional
        The upper and lower saturation thresholds.
    
    Returns
    -------
    ndarray
        The quantized array.
    """
    
    # check validity of quantizer parameters
    if step < 0:
        raise ValueError("The step must be positive.")
    if thresholds is not None:
        if len(thresholds) != 2:
            raise ValueError("Both upper and lower thresholds must be given.")
        if thresholds[0] >= thresholds[1]:
            raise ValueError("Invalid thresholds.")
    
    # perform quantization using the given step
    q = step*np.floor(x/step + 0.5)
    
    # saturate to the thresholds, if given
    if thresholds is not None:
        q[np.where(q < thresholds[0])] = thresholds[0]
        q[np.where(q > thresholds[1])] = thresholds[1]
    
    return q

def is_scalar(c):
    """
    Check if scalar.
    """
    
    return isinstance(c, numbers.Number)

def soft_thresholding(x, penalty):
    r"""
    Soft-thresholding.
    
    The function computes the element-wise soft-trhesholding defined as
    
        .. math:: \operatorname{sign}(x) \max\{ |x| - \rho, 0 \}
    
    where :math:`\rho` is a positive penalty parameter.
    
    Parameters
    ----------
    x : array_like
        Where to evaluate the soft-thresholding.
    penalty : float
        The positive penalty parameter :math:`\rho`.

    Returns
    -------
    ndarray
        The soft-thresolding of `x`.
    """
    
    return np.multiply(np.sign(x), np.maximum(np.absolute(x) - penalty, 0))

def bisection_method(f, a, b, tol=1e-5):
    """
    Minimize using the bisection method.
    
    This function minimizes a function `f` using the bisection method, stopping
    when :math:`a - b \leq t` for some threshold :math:`t`.

    Parameters
    ----------
    f
        The scalar function to be minimized.
    a : float
        The lower bound of the initial interval.
    b : float
        the upper bound of the initial interval.
    tol : float, optional
        The stopping condition, defaults to 1e-5.

    Returns
    -------
    x : float
        The approximate minimizer.
    """
    
    x = (a + b) / 2
    
    stop = False
    while not stop:
        
        f_x = f(x)
        
        if f(a)*f_x < 0: b = x
        elif f(b)*f_x < 0: a = x
        
        x = (a + b) / 2
        
        stop = abs(b - a) <= tol
    
    return x

def print_progress(i, num_iter, bar_length=80, decimals=2):
    """
    Print the progresso to command line.
    
    Parameters
    ----------
    i : int
        Current iteration.
    num_iter : int
        Total number of iterations.
    bar_length : int, optional
        Length of progress bar.
    decimals : int, optional
        Decimal places of the progress percent.
    
    Notes
    -----
    Adapted from `here 
    <https://gist.github.com/aubricus/f91fb55dc6ba5557fbab06119420dd6a>`_.
    """
    
    # completion percent
    str_format = "{0:." + str(decimals) + "f}"
    percents = str_format.format(100 * (i / num_iter))
    
    # progress bar
    filled_length = int(round(bar_length * i / num_iter))
    bar = "█" * filled_length + "-" * (bar_length - filled_length)
    
    # print progress
    sys.stdout.write("\rCompleted |%s| %s%s" % (bar, percents, "%")),
    if i == num_iter:
        sys.stdout.write("\n")
    sys.stdout.flush()

def initialize_trajectory(x_0, shape, num_iter):
    
    x = np.zeros(shape + (num_iter+1,))
    x[...,0] = x_0
    
    return x


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Distributed solvers.
"""


import numpy as np
from numpy.random import default_rng

ran = default_rng() # random generator

from tvopt import costs, solvers


#%% CONSENSUS METHODS

def average_consensus(net, x_0, num_iter=100):
    """
    Average consensus.
    
    Compute the average consensus over the network `net` with initial
    states `x_0`.

    Parameters
    ----------
    net : networks.Network
        The network describing the multi-agent system.
    x_0 : ndarray
        The initial states in a ndarray, with the last dimension indexing the 
        nodes.
    num_iter : int, optional
        The number of iterations to be performed.

    Returns
    -------
    x : ndarray
        The nodes' states after `num_iter` iterations.
    """
    
    x = x_0

    
    for l in range(num_iter): x = net.consensus(x)
    
    return x

def ratio_consensus(net, x_0, num_iter=100):
    """
    Ratio consensus.
    
    Compute the average consensus over the network `net` with initial
    states `x_0` using the ratio consensus protocol.

    Parameters
    ----------
    net : networks.Network
        The network describing the multi-agent system.
    x_0 : ndarray
        The initial states in a ndarray, with the last dimension indexing the 
        nodes.
    num_iter : int, optional
        The number of iterations to be performed.

    Returns
    -------
    x : ndarray
        The nodes' states after `num_iter` iterations.
    """
    
    x = x_0
    y, z = x, np.ones(x_0.shape) # auxiliary variables
    
    
    for l in range(num_iter):
        
        y_old, z_old = np.copy(y), np.copy(z)
        y[...], z[...] = 0, 0
        
        for i in range(net.N):
            for j in net.neighbors[i]+[i]:
                # y
                net.send(j, i, y_old[...,j]/(1+net.degrees[j]))
                y[...,i] += net.receive(i, j)
                # z
                net.send(j, i, z_old[...,j]/(1+net.degrees[j]))
                z[...,i] += net.receive(i, j)
        
        # perform consensus
        x = np.divide(y, z)
    
    return x

def gossip_consensus(net, x_0, num_iter=100, q=0.5):
    """
    Average consensus.
    
    Compute the average consensus over the network `net` with initial
    states `x_0` using the symmetric gossip protocol.

    Parameters
    ----------
    net : networks.Network
        The network describing the multi-agent system.
    x_0 : ndarray
        The initial states in a ndarray, with the last dimension indexing the 
        nodes.
    num_iter : int, optional
        The number of iterations to be performed.
    q : float, optional
        The weight used in the convex combination of the nodes that 
        communicate at each iteration.

    Returns
    -------
    x : ndarray
        The nodes' states after `num_iter` iterations.
    """
    
    x = x_0
    
    
    for l in range(num_iter):
                
        # choose nodes for gossip communication
        i = ran.integers(0, net.N)
        j = ran.choice(net.neighbors[i])
        
        # perform gossip communication
        net.send(i, j, x[...,i]), net.send(j, i, x[...,j])
        
        # update local states
        x[...,i] = q*x[...,i] + (1 - q)*net.receive(i, j)
        x[...,j] = q*x[...,j] + (1 - q)*net.receive(j, i)
    
    return x

def max_consensus(net, x_0, num_iter=100):
    """
    Max consensus.
    
    Compute the maximum of the nodes' states `x_0`.

    Parameters
    ----------
    net : networks.Network
        The network describing the multi-agent system.
    x_0 : ndarray
        The initial states in a ndarray, with the last dimension indexing the 
        nodes.
    num_iter : int, optional
        The number of iterations to be performed.

    Returns
    -------
    x : ndarray
        The nodes' states after `num_iter` iterations.
    """
    
    x = x_0
    
    
    for l in range(num_iter): x = net.max_consensus(x)
    
    return x


#%% PROXIMAL GRADIENT METHODS

def dpgm(problem, step, x_0=0, num_iter=100):
    r"""
    Distributed proximal gradient method (DPGM).
    
    This function implements the DPGM algorithm proposed in [1]_ (see also 
    [2]_ for the gradient-only version). The algorithm is characterized by 
    the following updates
    
    .. math:: \begin{align}
              \pmb{y}^\ell &= \pmb{W} \pmb{x}^\ell - \alpha \nabla f(\pmb{x}^\ell) \\
              \pmb{x}^{\ell+1} &= \operatorname{prox}_{\alpha g}(\pmb{y}^\ell)
              \end{align}
    
    for :math:`\ell = 0, 1, \ldots`. The algorithm is guaranteed to converge
    to a neighborhood of the optimal solution.

    Parameters
    ----------
    problem : dict
        A dictionary containing the network describing the multi-agent system
        and the costs describing the (possibly composite) problem. The 
        dictionary should contain :math:`f` and the network, and optionally
        :math:`g`.
    step : float
        The step-size.
    x_0 : ndarray, optional
        The initial states of the nodes. This can be either an ndarray
        of suitable size with the last dimension indexing the nodes, or
        a scalar. If it is a scalar then the same initial value is used for
        all components of the states.
    num_iter : int, optional
        The number of iterations to be performed.

    Returns
    -------
    x : ndarray
        The nodes' states after `num_iter` iterations.
    
    References
    ----------
    .. [1] Bastianello, N., Ajalloeian, A., & Dall'Anese, E. (2020). 
           Distributed and Inexact Proximal Gradient Method for Online Convex 
           Optimization. arXiv preprint arXiv:2001.00870.
    .. [2] Yuan, K., Ling, Q., & Yin, W. (2016). On the convergence of 
           decentralized gradient descent. SIAM Journal on Optimization, 
           26(3), 1835-1854.
    """
    
    # unpack problem data
    f, g, net = problem["f"], problem.get("g", None), problem["network"]
    is_composite = g is not None
    
    x = np.zeros(f.dom.shape)
    x[...] = x_0
    
    
    # run algorithm
    for l in range(num_iter):
        
        # gradient step
        y = net.consensus(x, net.weights) - step*f.gradient(x)

        # proximal step
        if is_composite: x = g.proximal(y, penalty=step)
        else: x = y
    
    return x

def pg_extra(problem, step, x_0=0, num_iter=100):
    r"""
    Proximal gradient exact first-order algorithm (PG-EXTRA).
    
    This function implements the PG-EXTRA algorithm proposed in [1]_ (see also 
    [2]_ for the gradient-only version, EXTRA). The algorithm is characterized 
    by the following updates
    
    .. math:: \begin{align}
              \pmb{y}^\ell &= \pmb{y}^{\ell-1} + \pmb{W} \pmb{x}^\ell
                           - \tilde{\pmb{W}} \pmb{x}^{\ell-1} 
                           - \alpha (\nabla f(\pmb{x}^\ell) - \nabla f(\pmb{x}^{\ell-1})) \\
              \pmb{x}^{\ell+1} &= \operatorname{prox}_{\alpha g}(\pmb{y}^\ell)
              \end{align}
    
    for :math:`\ell = 0, 1, \ldots`, where
    :math:`\tilde{\pmb{W}} = (\pmb{I} + \pmb{W}) /2`. The 
    algorithm is guaranteed to converge to the optimal solution.

    Parameters
    ----------
    problem : dict
        A dictionary containing the network describing the multi-agent system
        and the costs describing the (possibly composite) problem. The 
        dictionary should contain :math:`f` and the network, and optionally
        :math:`g`.
    step : float
        The step-size.
    x_0 : ndarray, optional
        The initial states of the nodes. This can be either an ndarray
        of suitable size with the last dimension indexing the nodes, or
        a scalar. If it is a scalar then the same initial value is used for
        all components of the states.
    num_iter : int, optional
        The number of iterations to be performed.

    Returns
    -------
    x : ndarray
        The nodes' states after `num_iter` iterations.
    
    References
    ----------
    .. [1] Shi, W., Ling, Q., Wu, G., & Yin, W. (2015). A proximal gradient 
            algorithm for decentralized composite optimization. IEEE 
            Transactions on Signal Processing, 63(22), 6013-6023.
    .. [2] Shi, W., Ling, Q., Wu, G., & Yin, W. (2015). Extra: An exact 
            first-order algorithm for decentralized consensus optimization. 
            SIAM Journal on Optimization, 25(2), 944-966.
    """
    
    # unpack problem data
    f, g, net = problem["f"], problem.get("g", None), problem["network"]
    is_composite = g is not None
    
    x = np.zeros(f.dom.shape)
    x[...] = x_0
    x_old = x
    
    # initialization
    x_cons = net.consensus(x, net.weights)
    grad = f.gradient(x)
    y = x_cons - step*grad
    
    if is_composite: x = g.proximal(y, penalty=step)
    else: x = y
    
    
    # run algorithm
    for l in range(1,num_iter):
        
        # store old consensus step and gradient evaluation
        x_cons_old, grad_old = np.copy(x_cons), np.copy(grad)
        
        # perform consensus step and gradient evaluation
        x_cons = net.consensus(x, net.weights)
        grad = f.gradient(x)
        
        # gradient tracking step
        y = y + x_cons - 0.5*(x_old + x_cons_old) - step*(grad - grad_old)
        x_old = x
        
        # proximal step
        if is_composite: x = g.proximal(y, penalty=step)
        else: x = y
    
    return x

def nids(problem, step, x_0=0, num_iter=100):
    r"""
    Network InDependent Step-size algorithm (NIDS).
    
    This function implements the NIDS algorithm proposed in [1]_. The algorithm
    is characterized by the following updates
    
    .. math:: \begin{align}
              \pmb{y}^\ell &= \pmb{y}^{\ell-1} - \pmb{x}^\ell
              - \tilde{\pmb{W}} (2 \pmb{x}^\ell - \pmb{x}^{\ell-1}
              - \operatorname{diag}(\pmb{\alpha}) (\nabla f(\pmb{x}^\ell) 
              - \nabla f(\pmb{x}^{\ell-1}))) \\
              \pmb{x}^{\ell+1} &= \operatorname{prox}_{\pmb{\alpha} g}(\pmb{y}^\ell)
              \end{align}
    
    for :math:`\ell = 0, 1, \ldots`, where :math:`\pmb{\alpha}` is a column
    vector containing the independent step-sizes of the nodes, and
    
    .. math:: \tilde{\pmb{W}} = \pmb{I} 
              + c \operatorname{diag}(\pmb{\alpha}) (\pmb{W} - \pmb{I})
    
    with :math:`c = 0.5 / \max_i \{ \alpha_i \}`. The algorithm is guaranteed
    to converge to the optimal solution.

    Parameters
    ----------
    problem : dict
        A dictionary containing the network describing the multi-agent system
        and the costs describing the (possibly composite) problem. The 
        dictionary should contain :math:`f` and the network, and optionally
        :math:`g`.
    step : float
        A common step-size or a list of local step-sizes, one for each node.
    x_0 : ndarray, optional
        The initial states of the nodes. This can be either an ndarray
        of suitable size with the last dimension indexing the nodes, or
        a scalar. If it is a scalar then the same initial value is used for
        all components of the states.
    num_iter : int, optional
        The number of iterations to be performed.

    Returns
    -------
    x : ndarray
        The nodes' states after `num_iter` iterations.
    
    References
    ----------
    .. [1] Li, Z., Shi, W., & Yan, M. (2019). A decentralized proximal-gradient
            method with network independent step-sizes and separated convergence
            rates. IEEE Transactions on Signal Processing, 67(17), 4494-4506.
    """

    # unpack problem data
    f, g, net = problem["f"], problem.get("g", None), problem["network"]
    is_composite = g is not None
    
    x = np.zeros(f.dom.shape)
    x[...] = x_0
    x_old = x
    
    # uncoordinated step-sizes
    step = step if isinstance(step, list) else [step for _ in range(net.N)]
    
    # compute consensus matrix
    common_step = 0.5 / np.max(step)
    
    weights = np.eye(net.N) + common_step*np.diag(step).dot(net.weights - np.eye(net.N))
    
    # initialization
    grad = f.gradient(x)
    y = x - np.stack([step[i]*grad[...,i] for i in range(net.N)], axis=-1)
    
    if is_composite: x = g.proximal(y, penalty=step)
    else: x = y
    
    
    # run algorithm
    for l in range(1,num_iter):
        
        # store old gradient and compute new one
        grad_old = np.copy(grad)
        grad = f.gradient(x)
        
        # gradient tracking step
        y = y - x + net.consensus(2*x - x_old \
          - np.stack([step[i]*(grad[...,i]-grad_old[...,i]) for i in range(net.N)], axis=-1), weights=weights)
        x_old = x
        
        # proximal step
        if is_composite: x = g.proximal(y, penalty=step)
        else: x = y
    
    return x

def prox_ed(problem, step, x_0=0, num_iter=100):
    r"""
    Proximal exact diffusion (Prox-ED).
    
    This function implements the Prox-ED algorithm [1]_. The algorithm is 
    characterized by the following updates
    
    .. math:: \begin{align}
              \pmb{y}^\ell &= \pmb{x}^\ell - \alpha \nabla f(\pmb{x}^\ell) \\
              \pmb{u}^\ell &= \pmb{z}^{\ell-1} + \pmb{y}^\ell - \pmb{y}^{\ell-1} \\
              \pmb{z}^\ell &= \tilde{\pmb{W}} \pmb{u}^\ell \\
              \pmb{x}^{\ell+1} &= \operatorname{prox}_{\alpha g}(\pmb{z}^\ell)
              \end{align}
    
    for :math:`\ell = 0, 1, \ldots`, where
    :math:`\tilde{\pmb{W}} = (\pmb{I} + \pmb{W}) /2`. The 
    algorithm is guaranteed to converge to the optimal solution.

    Parameters
    ----------
    problem : dict
        A dictionary containing the network describing the multi-agent system
        and the costs describing the (possibly composite) problem. The 
        dictionary should contain :math:`f` and the network, and optionally
        :math:`g`.
    step : float
        The step-size.
    x_0 : ndarray, optional
        The initial states of the nodes. This can be either an ndarray
        of suitable size with the last dimension indexing the nodes, or
        a scalar. If it is a scalar then the same initial value is used for
        all components of the states.
    num_iter : int, optional
        The number of iterations to be performed.

    Returns
    -------
    x : ndarray
        The nodes' states after `num_iter` iterations.
    
    References
    ----------
    .. [1] S. A. Alghunaim, E. Ryu, K. Yuan, and A. H. Sayed, "Decentralized 
           Proximal Gradient Algorithms with Linear Convergence Rates," IEEE 
           Transactions on Automatic Control, 2020.
    """
    
    # unpack problem data
    f, g, net = problem["f"], problem.get("g", None), problem["network"]
    is_composite = g is not None
    
    x = np.zeros(f.dom.shape)
    x[...] = x_0
    
    # consensus matrix
    weights = 0.5*(np.eye(net.N) + net.weights)
    
    # initialization
    y = x - step*f.gradient(x)
    u = np.copy(y)
    z = net.consensus(u, weights=weights)
    
    if is_composite: x = g.proximal(z, penalty=step)
    else: x = z
    
    
    # run algorithm
    for l in range(1,num_iter):
        
        # store old y variables
        y_old = np.copy(y)
        
        # gradient updates
        y = x - step*f.gradient(x)
        u = z + y - y_old
        
        # communication step
        z = net.consensus(u, weights=weights)
        
        # proximal step
        if is_composite: x = g.proximal(z, penalty=step)
        else: x = z
    
    return x


#%% GRADIENT METHODS

def aug_dgm(problem, step, x_0=0, num_iter=100):
    r"""
    Augmented distributed gradient method (Aug-DGM).
    
    This function implements the Aug-DGM algorithm (see [1]_). The algorithm is
    characterized by the following updates
    
    .. math:: \begin{align}
              \pmb{y}^\ell &= \pmb{W} \left( \pmb{y}^{\ell-1} 
                            + \nabla f(\pmb{x}^\ell) - \nabla f(\pmb{x}^{\ell-1}) \right) \\
              \pmb{x}^{\ell+1} &= \pmb{W} \left( \pmb{x}^\ell - \pmb{A} \pmb{y}^\ell \right)
              \end{align}
    
    for :math:`\ell = 0, 1, \ldots` where :math:`\pmb{A}` is a diagonal matrix 
    of uncoordinated step-sizes. The algorithm is guaranteed to converge 
    to the optimal solution.

    Parameters
    ----------
    problem : dict
        A dictionary containing the network describing the multi-agent system
        and the cost describing the problem.
    step : float
        A common step-size or a list of local step-sizes, one for each node.
    x_0 : ndarray, optional
        The initial states of the nodes. This can be either an ndarray
        of suitable size with the last dimension indexing the nodes, or
        a scalar. If it is a scalar then the same initial value is used for
        all components of the states.
    num_iter : int, optional
        The number of iterations to be performed.

    Returns
    -------
    x : ndarray
        The nodes' states after `num_iter` iterations.
    
    References
    ----------
    .. [1] J. Xu, S. Zhu, Y. C. Soh, and L. Xie, "Augmented distributed 
           gradient methods for multi-agent optimization under uncoordinated 
           constant stepsizes," in 2015 54th IEEE Conference on Decision and 
           Control (CDC), Osaka, Japan, Dec. 2015, pp. 2055–2060.
    """

    # unpack problem data
    f, net = problem["f"], problem["network"]
    
    x = np.zeros(f.dom.shape)
    x[...] = x_0
    
    # uncoordinated step-sizes
    step = step if isinstance(step, list) else [step for _ in range(net.N)]
    
    # initialization
    grad = f.gradient(x)
    y = grad
    x = net.consensus(x - np.stack([step[i]*y[...,i] \
               for i in range(net.N)], axis=-1), net.weights)
    
    
    # run algorithm
    for l in range(1,num_iter):
        
        # old and new gradient evaluations
        grad_old = np.copy(grad)
        grad = f.gradient(x)
        
        # gradient tracking step
        y = net.consensus(y + grad - grad_old, net.weights)
        x = net.consensus(x - np.stack([step[i]*y[...,i] \
                     for i in range(net.N)], axis=-1), net.weights)
    
    return x


#%% DUAL-BASED METHODS

def dual_ascent(problem, step, w_0=0, num_iter=100):
    r"""
    Distributed dual ascent a.k.a. dual decomposition (DD).
    
    This function implements the DD algorithm [1]_. The algorithm is 
    characterized by the following updates
    
    .. math:: \pmb{x}^\ell = \operatorname{arg\,min}_{\pmb{x}} \left\{
        f(\pmb{x}) - \langle (\pmb{I} - \pmb{W}) \pmb{x}, \pmb{w}^\ell 
        \rangle\right\}
    
    .. math:: \pmb{w}^{\ell+1} = \pmb{w}^\ell 
                                - \alpha (\pmb{I} - \pmb{W}) \pmb{x}^\ell
    
    for :math:`\ell = 0, 1, \ldots`, where :math:`\pmb{w}` is the vector of
    Lagrange multipliers. The algorithm is guaranteed to converge to the 
    optimal solution.

    Parameters
    ----------
    A dictionary containing the network describing the multi-agent system
        and the cost describing the problem.
    step : float
        The step-size.
    w_0 : ndarray, optional
        The initial value of the dual nodes' states. This can be either an
        ndarray of suitable size with the last dimension indexing the nodes, or
        a scalar. If it is a scalar then the same initial value is used for
        all components.
    num_iter : int, optional
        The number of iterations to be performed.

    Returns
    -------
    x : ndarray
        The nodes' states after `num_iter` iterations.
    w : ndarray
        The dual variables of the nodes after `num_iter` iterations.
    
    References
    ----------
    .. [1] Simonetto, A. (2018). Dual Prediction–Correction Methods for
            Linearly Constrained Time-Varying Convex Programs. IEEE Transactions
            on Automatic Control, 64(8), 3355-3361.
    """
    
    # unpack problem data
    f, net = problem["f"], problem["network"]
    
    x = np.zeros(f.dom.shape)
    w = np.zeros(f.dom.shape) # dual variables
    w[...] = w_0

    
    # run algorithm
    for l in range(num_iter):
        
        # dual communication step
        w_comm = w - net.consensus(w, net.weights)
        
        # primal update
        for i in range(net.N):
            x[...,i] = solvers.newton({"f":f.costs[i]+costs.Linear(-w_comm[...,i])}, x_0=x[...,i], tol=1e-5)
        
        # primal communication step
        x_comm = x - net.consensus(x, net.weights)
        
        # dual update
        w = w - step*x_comm
    
    return x, w

def admm(problem, penalty, rel, w_0=0, num_iter=100):
    r"""
    Distributed relaxed alternating direction method of multipliers (ADMM).
    
    This function implements the distributed ADMM, see [1]_ and references 
    therein. The algorithm is characterized by the following updates
    
    .. math:: x_i^\ell = \operatorname{prox}_{f_i / (\rho d_i)}
                                    ([\pmb{A}^\top z^\ell]_i / (\rho d_i))
    
    .. math:: z_{ij}^{\ell+1} = (1-\alpha) z_{ij}^\ell - \alpha z_{ji}^\ell 
                              + 2 \alpha \rho x_j^\ell
    
    for :math:`\ell = 0, 1, \ldots`, where :math:`d_i` is node :math:`i`'s 
    degree, :math:`\rho` and :math:`\alpha` are the penalty and relaxation
    parameters, and :math:`\pmb{A}` is the arc incidence matrix. The algorithm 
    is guaranteed to converge to the optimal solution.

    Parameters
    ----------
    problem : dict
        A dictionary containing the network describing the multi-agent system
        and the cost describing the problem.
    penalty : float
        The penalty parameter :math:`\rho` of the algorithm (convergence is
        guaranteed for any positive value).
    rel : float
        The relaxation parameter :math:`\alpha` of the algorithm (convergence 
        is guaranteed for values in :math:`(0,1)`).
    w_0 : ndarray, optional
        The initial value of the dual nodes' states. This can be either an
        ndarray of suitable size with the last dimension indexing the nodes, or
        a scalar. If it is a scalar then the same initial value is used for
        all components.
    num_iter : int, optional
        The number of iterations to be performed.

    Returns
    -------
    x : ndarray
        The nodes' states after `num_iter` iterations.
    w : ndarray
        The dual variables of the nodes after `num_iter` iterations.
    
    References
    ----------
    .. [1] N. Bastianello, R. Carli, L. Schenato, and M. Todescato, 
           "Asynchronous Distributed Optimization over Lossy Networks via 
           Relaxed ADMM: Stability and Linear Convergence," IEEE Transactions 
           on Automatic Control.
    """
    
    # unpack problem data
    f, net = problem["f"], problem["network"]
    
    x = np.zeros(f.dom.shape)
    
    # parameters for local proximal evaluations
    penalty_i = [1 / (penalty*net.degrees[i]) for i in range(net.N)]
    
    # initialize arc variables
    z = {}
    for i in range(net.N):
        for j in net.neighbors[i]:
            z[i,j] = np.zeros(f.dom.shape[:-1])
    
    
    for l in range(num_iter):
        
        for i in range(net.N):
            
            # primal update
            x[...,i] = f.proximal(sum([z[i,j] for j in net.neighbors[i]]) * penalty_i[i], penalty=penalty_i[i], i=i)
                
            # communication step
            for j in net.neighbors[i]: net.send(i, j, -z[i,j] + 2*penalty*x[...,i])
        
        # update auxiliary variables
        for i in range(net.N):
            for j in net.neighbors[i]:
                
                z[i,j] = (1 - rel)*z[i,j] + rel*net.receive(i, j)
    
    
    return x, z


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Network tools.
"""

import math
import numpy as np
from numpy.random import default_rng
from numpy import linalg as la
from functools import partialmethod

ran = default_rng() # random generator

from tvopt import sets, utils


#%% NETWORK CLASSES

class Network():
    """
    Representation of an undirected network.
    
    The class implements an undirected network defined from the adjacency
    matrix. The class provides methods for different communication protocols,
    such as node-to-node and broadcast.
    
    Transmissions are implemented via the `buffer` attribute of the network:
    the sender stores the packet to be transmitted in the `buffer` dictionary,
    specifying the recipient, which can then access the packet.
    
    By convention, the nodes in the network are indexed from :math:`0` to
    :math:`N-1`, where :math:`N` is the total number of nodes.
    
    Attributes
    ----------
    adj_mat : ndarray
        The adjacency matrix of the network.
    N : ndarray
        The number of nodes in the network.
    weights : ndarray
        The consensus weight matrix, if not specified in the constructor
        this is the Metropolis-Hastings weight matrix.
    neighbors : list
        A list whose :math:`i`-th element is a list of node :math:`i`'s
        neighors.
    degrees : list
        The number of neighbors of each node.
    buffer : dict
        The dictionary used for node-to-node transmissions.
    """
    
    def __init__(self, adj_mat, weights=None):
        """
        Class constructor.

        Parameters
        ----------
        adj_mat : ndarray
            The adjacency matrix describing the connectivity pattern of the
            network.
        weights : ndarray, optional
            The consensus matrix used by the network. If the argument is not
            specified, then a Metropolis-Hastings weight matrix is generated.
        """
        
        # store network attributes
        self.adj_mat, self.N = adj_mat, adj_mat.shape[0]
        
        # consensus matrix
        self.weights = weights if weights is not None else metropolis_hastings(self.adj_mat)
        
        # list of neighbors for each node
        self.neighbors = [list(np.where(self.adj_mat[i,:])[0]) for i in range(self.N)]
        self.degrees = [len(n) for n in self.neighbors]
        
        # buffer for node-to-node transmissions
        self.buffer = {}
    
    def send(self, sender, receiver, packet):
        """
        Node-to-node transmission (sender phase).
        
        This method simulates a node-to-node transmission by storing the packet
        to be communicated in the `buffer`. In particular, if :math:`i` is the
        sender and :math:`j` the receiver, then the packet is introduced in
        `buffer` with keyword :math:`(j,i)`.
        
        Note that older information (if any) in the `buffer` is overwritten 
        whenever `send` is called.

        Parameters
        ----------
        sender : int
            The index of the transmitting node.
        receiver : int
            The index of the recipient.
        packet : array_like
            The packet to ne communicated.
        """
        
        self.buffer[receiver, sender] = packet
    
    def receive(self, receiver, sender, default=0, destructive=True):
        """
        Node-to-node transmission (receiver phase).
        
        This method simulates the reception of a packet previously transmitted
        using the `send` method. In patricular, the method accesses the 
        packet in the `buffer` dictionary. If the packet is not present,
        a default value is returned.
        
        Reads from the `buffer` can be destructive, meaning that the packet
        is read and removed, which is the default, or not.

        Parameters
        ----------
        receiver : int
            The index of the recipient.
        sender : int
            The index of the transmitting node.
        default : array_like, optional
            The value returned when a packet from `sender` to `receiver` is not
            found in the `buffer`.
        destructive : bool, optional
            Specifies if the packet should be removed from the `buffer` after
            being read (which is the default) or not.

        Returns
        -------
        array_like
            The packet or a default value.
        """
        
        return self.buffer.pop((receiver, sender), default) if destructive \
               else self.buffer.get((receiver, sender), default)
    
    def broadcast(self, sender, packet):
        """
        Broadcast transmission.
        
        This method implements a broadcast transmission in which a node sends
        the same packet to all its neighbors. The packet is also transmitted
        to the node itself. The method is implemented using the `send` method.

        Parameters
        ----------
        sender : int
            The index of the transmitting node.
        packet : array_like
            The packet to ne communicated.
        """
        
        for r in self.neighbors[sender]+[sender]: self.send(sender, r, packet)
    
    def consensus(self, x, weights=None):
        """
        Consensus mixing.
        
        This method implements a consensus step over the network, mixing the 
        given nodes' states using the weight matrix of the network or a 
        different one.

        Parameters
        ----------
        x : array_like
            The nodes' local states in an array with the last dimension 
            indexing the nodes.
        weights : ndarray, optional
            The consensus weight matrix to be used instead of the one created
            at initialization.

        Returns
        -------
        y : ndarray
            The local states after a consensus step.
        """
        
        # consensus matrix
        weights = weights if weights is not None else self.weights
        
        # broadcast the states
        for i in range(self.N): self.broadcast(i, x[...,i])
        
        # perform consensus
        y = np.empty(x.shape)
        for i in range(self.N):
            y[...,i] = sum([weights[i,j]*self.receive(i, j) for j in self.neighbors[i]+[i]])
                    
        return y
    
    def max_consensus(self, x):
        """
        Max-consensus.
        
        This method implements a step of max-consensus, where each node selects
        the (element-wise) maximum between the packets received from its 
        neighbors and its own state. See [1]__ for a reference on 
        max-consensus.

        Parameters
        ----------
        x : array_like
            The nodes' local states in an array with the last dimension 
            indexing the nodes.

        Returns
        -------
        x : ndarray
            The local states after a max-consensus step.
        
        References
        ----------
        .. [1] F. Iutzeler, P. Ciblat, and J. Jakubowicz, "Analysis of 
               Max-Consensus Algorithms in Wireless Channels," IEEE 
               Transactions on Signal Processing, vol. 60, no. 11, pp. 
               6103–6107, Nov. 2012.
        """
        
        # broadcast the states
        for i in range(self.N): self.broadcast(i, x[...,i])
        
        # perform max consensus
        for i in range(self.N): x[...,i] = np.max([self.receive(i, j, -np.inf) for j in self.neighbors[i]+[i]], axis=0)
        
        return x


# -------- EXAMPLES

class NoisyNetwork(Network):
    """
    Network with Gaussian communication noise.
    
    Representation of a connected, undirected network, whose communication
    protocol is subject to additive white Gaussian noise. The network's
    transmission methods add normal noise to all packets (unless they are sent
    from a node to itself).
    """
    
    def __init__(self, adj_mat, noise_var, weights=None):
        """
        Class constructor from adjacency matrix.
        
        Parameters
        ----------
        adj_mat : ndarray
            Adjacency matrix of the given graph.
        n : int
            The size of local states.
        noise_var : float
            The white Gaussian noise variance.
        weights : ndarray, optional
            The edge weights of the graph. If None, Metropolis-Hasting weights
            are generated by default.
        
        Raises
        ------
        ValueError.
        """   
        
        # call the super-class constructor
        super().__init__(adj_mat, weights)        
        self.std = math.sqrt(noise_var)

    def send(self, sender, receiver, packet):
        
        # generate noise
        noise = self.std*ran.standard_normal() if sender != receiver else 0
        # send noisy packet
        self.buffer[receiver, sender] = packet + noise

class QuantizedNetwork(Network):
    """
    Network with quantized communications.
    
    Representation of a connected, undirected network, whose communications
    are quantized. The network's transmission methods quantize all packets 
    (unless they are sent from a node to itself).
    """
    
    def __init__(self, adj_mat, step, thresholds=None, weights=None):
        """
        Class constructor from adjacency matrix.
        
        This object simulates a network that uses a quantized communication
        protocol. The quantization is uniform with the specified step, and
        optionally saturates to the given lower and upper thresholds.
        
        Parameters
        ----------
        adj_mat : ndarray
            Adjacency matrix of the given graph.
        n : int
            The size of local states.
        step : float
            The step of the quantizer.
        thresholds : list, optional
            The lower and upper saturation thresholds of the quantizer.
        weights : ndarray, optional
            The edge weights of the graph. If None, Metropolis-Hasting weights
            are generated by default.
        """   
        
        # call the super-class constructor
        super().__init__(adj_mat, weights)
        
        # store the quantizer parameters (their validity is implicitly checked 
        # when calling utils.uniform_quantizer)
        self.step, self.thresholds = step, thresholds
        
    def send(self, sender, receiver, packet):
        
        # generate noise
        q = utils.uniform_quantizer(packet, self.step, self.thresholds) if sender != receiver else packet
        # send quantized packet
        self.buffer[receiver, sender] = q

class LossyNetwork(Network):
    """
    Network with random communication failures.
    
    Representation of a connected, undirected network, whose communication
    protocol is subject to packet losses. Packet sent from a node to another
    may be lost with a certain probability.
    """
    
    def __init__(self, adj_mat, loss_prob, weights=None): 
        
        # call the super-class constructor
        super().__init__(adj_mat, weights)
        
        # store the communication noise variance
        if loss_prob < 0 or loss_prob >= 1:
            raise ValueError("The packet loss probability should be in [0,1).")
        
        self.loss_prob = loss_prob
    
    def send(self, sender, receiver, packet):
        # the packet always arrives if the sender==receiver, or it arrives w.p. 1-loss_prob
        
        # try to send the packet
        if sender == receiver or ran.random() > self.loss_prob: self.buffer[receiver, sender] = packet


# -------- EXAMPLES

class DynamicNetwork(Network):
    """
    Time-varying network.
    
    This class creates a time-varying network from a list of network objects,
    and possibly a sampling time that specifies how often the network changes.
    """
    
    def __init__(self, nets, t_s=1):
        """
        Class constructor.

        Parameters
        ----------
        nets : list
            A list of the networks over time.
        t_s : float, optional
            The time after which the network changes to the next configuration.
        """
            
        # create time domain
        self.time = sets.T(t_s, t_min=0, t_max=len(nets)*t_s)
        # store networks
        self.nets, self.N = nets, len(nets)
    
    def _evaluate(self, method, t, *args, **kwargs):
        
        return getattr(self.nets[self.time.check_input(t)], method)(*args, **kwargs)
    
    send = partialmethod(_evaluate, "send")
    broadcast = partialmethod(_evaluate, "broadcast")
    consensus = partialmethod(_evaluate, "consensus")
    max_consensus = partialmethod(_evaluate, "max_consensus")

    def sample(self, t):
        """
        Sample the dynamic network.
        
        This method returns the network object that is active at time `t`.

        Parameters
        ----------
        t : float
            The time when the network should be sampled.

        Returns
        -------
        Network
            The sampled network.
        """
        
        return self.nets[self.time.check_input(t)]


#%% GRAPH GENERATION

def random_graph(N, radius):
    """
    Generate a random geometric graph.
    
    Parameters
    ----------
    N : int
        Number of nodes in the graph.
    radius : float
        Radius of each node's neighborhood, must be in :math:`[0,1)`.
    
    Returns
    -------
    adj_mat : ndarray
        Adjacency matrix of the generated graph.
    
    Raises
    ------
    ValueError.
    
    Notes
    -----
    The function recursively generates random positions for the nodes on the
    :math:`[0,1] \\times [0,1]` square, and then builds the graph by setting 
    as neighbors each pair of nodes within a distance no larger than `radius`. 
    The process is repeated until the result is a connected graph. For this 
    reason, combinations of small `N` and `radius` *can yield 
    exceedingly long computation times*. If the computation does not succeed
    after `2500` iterations, an error is raised.
    """
    
    # check validity of arguments
    if N <= 0:
        raise ValueError("The number of nodes must be positive.")   
    if radius < 0 or radius > 1:
        raise ValueError("Radius must be a number in [0,1).")

    # (try to) cast N as integer
    N = int(N)
    
    # stopping conditions
    connected, max_iter = False, 2500
    
    i = 0
    while not connected and i < max_iter:
        
        # generate nodes' positions
        pos = ran.random((N, 2)) # random positions on the plane
        
        dist_x = np.repeat(pos[:,0], N).reshape((N,N)) \
               - np.repeat(pos[:,0], N).reshape((N,N)).T
        dist_y = np.repeat(pos[:,1], N).reshape((N,N)) \
               - np.repeat(pos[:,1], N).reshape((N,N)).T
        
        dist = np.square(dist_x) + np.square(dist_y) # nodes' distances
        
        # build adjacency matrix
        adj_mat = (dist < radius).astype(float)
        # and remove self-loops
        adj_mat = adj_mat - np.diag(np.diag(adj_mat))

        # check if graph is connected        
        connected = is_connected(adj_mat)
        i += 1
    
    # return graph
    if connected:
        return adj_mat - np.diag(np.diag(adj_mat))
    else:
        raise ValueError("Unable to generate the graph, try a larger radius.")

def erdos_renyi(N, prob):
    """
    Generate a random Erdos-Renyi graph.

    Parameters
    ----------
    N : int
        Number of nodes in the graph.
    prob : float
        The probability of adding an edge between any two nodes.

    Returns
    -------
    adj_mat : ndarray
        Adjacency matrix of the generated graph.
    
    Raises
    ------
    ValueError.
    """
    
    # check validity of arguments
    if N <= 0:
        raise ValueError("The number of nodes must be positive.")   
    if prob <= 0 or prob > 1:
        raise ValueError("`prob` must be a number in (0,1].")

    # (try to) cast N as integer
    N = int(N)
    
    # stopping conditions
    connected, max_iter = False, 2500
    
    i = 0
    while not connected and i < max_iter:
        
        # generate the edges
        adj_mat = np.zeros((N, N))
        
        for i in range(N):
            for j in range(i+1, N):
                                
                if ran.random() <= prob: adj_mat[i,j], adj_mat[j,i] = 1, 1

        # check if graph is connected        
        connected = is_connected(adj_mat)
        i += 1
    
    # return graph
    if connected:
        return adj_mat - np.diag(np.diag(adj_mat))
    else:
        raise ValueError("Unable to generate the graph, try a larger probability.")


def circulant_graph(N, num_conn):
    """
    Generate a circulant graph.
    
    Parameters
    ----------
    N : int
        Number of nodes in the graph.
    num_conn : int
        Number of neighbors on each side of a node.
    
    Returns
    -------
    adj_mat : ndarray
        Adjacency matrix of the generated graph.
    
    Notes
    -----
    If `num_conn` is larger than `N` / 2 a complete graph is returned.
    """
    
    # check validity of arguments
    if N <= 0 or num_conn <= 0:
        raise ValueError("The number of nodes and connections must be positive.")   
        
    # (try to) cast N and num_conn as integers
    N = int(N)
    num_conn = int(num_conn)
    
    adj_mat = np.zeros((N,N))
    
    # create base vector
    if 2*num_conn+1 >= N:
        vec = np.hstack((0, np.ones(N-1)))
    else:
        vec = np.hstack((0, np.ones(num_conn), 
                         np.zeros(N-2*num_conn-1), np.ones(num_conn)))
    
    for i in range(N):
        
        adj_mat[i,:] = np.roll(vec, i)
            
    return adj_mat


def circle_graph(N):
    """
    Generate a circle graph.
    
    Parameters
    ----------
    N : int
        Number of nodes in the graph.
    
    Returns
    -------
    adj_mat : ndarray
        Adjacency matrix of the generated graph.
    
    See Also
    --------
    circulant_graph : Circulant graph generator
    """
    
    return circulant_graph(N, 1)


def complete_graph(N):
    """
    Generate a complete graph.
    
    Parameters
    ----------
    N : int
        Number of nodes in the graph.
    
    Returns
    -------
    adj_mat : ndarray
        Adjacency matrix of the generated graph.
    
    See Also
    --------
    circulant_graph : Circulant graph generator
    """
    
    return circulant_graph(N, N)
    

def star_graph(N):
    """
    Generate a star graph.
    
    Parameters
    ----------
    N : int
        Number of nodes in the graph.
    
    Returns
    -------
    adj_mat : ndarray
        Adjacency matrix of the generated graph.
    """
    
    # check validity of arguments
    if N <= 0:
        raise ValueError("The number of nodes must be positive.")   
        
    # (try to) cast N as integer
    N = int(N)
    
    adj_mat = np.zeros((N,N))
    adj_mat[0,1:N] = 1
    adj_mat[1:N,0] = 1
            
    return adj_mat


#%% GRAPH MATRICES

def metropolis_hastings(adj_mat):
    """ 
    Compute a consensus matrix based on the Metropolis-Hastings rule.
    
    The Metropolis-Hastings rule generates a matrix :math:`W` with off-diagonal
    elements equal to:
        
        .. math:: w_{ij} = \\frac{1}{1 + \max\{ d_i, d_j \}}
        
    where :math:`i` is a node index and :math:`j \\neq i` the index of one of 
    its neighbors, and :math:`d_i`, :math:`d_j` are their respective degrees. 
    The diagonal elements are assigned as:
    
        .. math:: w_{ii} = 1 - \sum_{j \in \mathcal{N}_i} w_{ij}
    
    to guarantee double stochasticity.
    
    Parameters
    ----------
    adj_mat : ndarray
        Adjacency matrix describing the graph.
    
    Returns
    -------
    mh_mat : ndarray
        Metropolois-Hastings consensus matrix.
    """
    
    # retrieve num. of nodes
    N = adj_mat.shape[0]
    # nodes' degrees
    deg = np.sum(adj_mat,axis=0)
    
    mh_mat = np.zeros((N,N))
    
    # add the off-diagonal elements
    for i in range(N):
        for j in np.where(adj_mat[i,:])[0]:
            
            mh_mat[i,j] = 1 / (1 + np.max((deg[i],deg[j])))
    
    # add the diagonal elements
    mh_mat = mh_mat + np.eye(N) - np.diag(np.sum(mh_mat,axis=0))
    
    return mh_mat

def incidence_matrix(adj_mat, n=1):
    """
    Build the incidence matrix.
    
    The edges :math:`e = (i,j)` are ordered with :math:`i \leq j`, so that 
    in the :math:`e`-th column the :math:`i`-th element is :math:`1` and
    the :math:`j`-th is :math:`-1` (the remaining are of course :math:`0`).
    
    Parameters
    ----------
    adj_mat : ndarray
        Adjacency matrix describing the graph.
    n : int, optional
        Size of the local states.

    Returns
    -------
    incid_mat : ndarray
        The incidence matrix.
    """
    
    # return upper triangular portion [not to count edges twice]
    adj_mat = np.triu(adj_mat)

    N = adj_mat.shape[0] # num of nodes 
    num_edges = int(np.sum(adj_mat)) # num of edges
    
    incid_mat = np.zeros((num_edges,N))
    
    # index of current edge
    e = 0
    # fill in the incidence matrix
    for i in range(N):
        for j in np.where(adj_mat[i,:])[0]:
            
            incid_mat[e,i] = 1
            incid_mat[e,j] = -1
            
            e += 1
            
    return np.kron(incid_mat, np.eye(n))


#%% TOOLS

def is_connected(adj_mat):
    """
    Verify if a graph is connected.
    
    Parameters
    ----------
    adj_mat : ndarray
        Adjacency matrix describing the graph.

    Returns
    -------
    bool
        True if the graph is connected, False otherwise.
    
    Notes
    -----
    The connectedness of the graph is checked by verifying whether the
    :math:`N`-th power of the adjacency matrix plus the identity is a full 
    matrix (no zero elements), with :math:`N` the number of nodes.
    """
    
    N = adj_mat.shape[0]
    
    return np.all(la.matrix_power(adj_mat + np.eye(N), N))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tvopt", # Replace with your own username
    version="0.1.1",
    author="Nicola Bastianello",
    author_email="nicola.bastianello.3@phd.unipd.it",
    description="tvopt: A Python Framework for Time-Varying Optimization",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nicola-bastianello/tvopt",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy>=1.19',
        'scipy>=1.5'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Cost prediction tools.
"""

import numpy as np

from tvopt import costs


#%% PREDICTION TEMPLATE

class Prediction(costs.Cost):
    """
    Prediction of a dynamic cost.
    
    This class creates a cost object that predicts a given dynamic function.
    The object stores a dynamic cost and a predicted cost, which can be 
    modified using new information through the method `update`.
    """
    
    def __init__(self, cost):
        
        super().__init__(cost.dom)
        self.cost, self.t_s = cost, cost.time.t_s
        
        # variable to store the last time the prediction was updated
        self.t_k = None
    
    def update(self, t, *args, **kwargs):
        """
        Update the current prediction.
        
        This method updates the current prediction by building a new predicted
        cost using the samples observed up to time `t`. By default this method 
        samples the dynamic cost, and should be overwritten when implementing 
        a custom prediction strategy.

        Parameters
        ----------
        t : float
            The time of the last sampled cost.
        """
        
        if self.t_k != t:
            self.prediction = self.cost.sample(t)
            self.t_k = t

    def function(self, x, **kwargs):
        
        return self.prediction.function(x, **kwargs)
    
    def gradient(self, x, **kwargs):
        
        return self.prediction.gradient(x, **kwargs)
    
    def hessian(self, x, **kwargs):
        
        return self.prediction.hessian(x, **kwargs)
    
    def proximal(self, x, penalty=1, **kwargs):
        
        return self.prediction.proximal(x, penalty=penalty, **kwargs)


#%% EXAMPLES

class ExtrapolationPrediction(Prediction):
    r"""
    Extrapolation-based prediction.
    
    This prediction strategy, proposed in [1]__, predicts the cost at time
    :math:`t_{k+1}` as:
    
        .. math:: \hat{f}(\pmb{x};t_{k+1}) = \sum_{i = 1}^I \ell_i f(\pmb{x}; t_{k - i + 1})
    
    where :math:`I \in \mathbb{N}` denotes the order, that is, the number of
    past functions to use, and with coefficients:
        
        .. math:: \ell_i = \prod_{1 \leq j \leq I, \ j \neq i} \frac{j}{j - i}.
        
    .. [1] N. Bastianello, A. Simonetto, and R. Carli, "Primal and Dual 
           Prediction-Correction Methods for Time-Varying Convex Optimization,"
           arXiv:2004.11709 [cs, math], Oct. 2020. Available:
           http://arxiv.org/abs/2004.11709.
    """
    
    def __init__(self, cost, order=2):
        
        super().__init__(cost)
        self.smooth = 2
        
        # extrapolation data
        self.order, self.range = order, list(range(1, order+1))
        self.coeffs = {i : np.prod([j / (j - i) for j in self.range[:i-1] + self.range[i:]]) for i in self.range}
    
    def update(self, t):
        
        # update if the t_k is different from the last used
        if self.t_k != t:
            
            if t < self.order*self.t_s: self.prediction = self.cost.sample(t)
            else: self.prediction = sum([self.coeffs[i]*self.cost.sample(t-(i-1)*self.t_s) for i in self.range])
            
            self.t_k = t

class TaylorPrediction(Prediction):
    r"""
    Taylor expansion-based prediction.
    
    This prediction strategy, proposed in [1]__ and see also [2]__, predicts 
    the cost at time :math:`t_{k+1}` using its Taylor expansion around 
    :math:`t_k` and a given :math:`\pmb{x}_k`:
        
        .. math:: \begin{split} \hat{f}(\pmb{x}; t_{k+1}) &= f(\pmb{x}_k;t_k) + \langle \nabla_x f(\pmb{x}_k;t_k), \pmb{x} - \pmb{x}_k \rangle
                                            + T_s \nabla_t f(\pmb{x}_k;t_k) + (T_s^2 / 2) \nabla_{tt} f(\pmb{x}_k;t_k) +\\
                                            &+ T_s \langle \nabla_{tx} f(\pmb{x}_k;t_k), \pmb{x} - \pmb{x}_k \rangle 
                                            + \frac{1}{2} (\pmb{x} - \pmb{x}_k)^\top \nabla_{xx} f(\pmb{x}_k;t_k) (\pmb{x} - \pmb{x}_k)
                                            \end{split}
    
    where :math:`T_s` is the sampling time.
        
    
    References
    ----------
    .. [1] A. Simonetto, A. Mokhtari, A. Koppel, G. Leus, and A. Ribeiro, 
           "A Class of Prediction-Correction Methods for Time-Varying 
           Convex Optimization," IEEE Transactions on Signal Processing, 
           vol. 64, no. 17, pp. 4576–4591, Sep. 2016.
    .. [2] N. Bastianello, A. Simonetto, and R. Carli, "Primal and Dual 
           Prediction-Correction Methods for Time-Varying Convex Optimization,"
           arXiv:2004.11709 [cs, math], Oct. 2020. Available:
           http://arxiv.org/abs/2004.11709.
    """
    
    def __init__(self, cost):
        
        super().__init__(cost)
        self.smooth = 2
        
        # variables for where the last prediction was centered
        self.x_k = None
    
    def update(self, t, x, gradient_only=True, **kwargs):
        
        # update if the x_k or t_k are different from the last used
        if np.any(self.x_k != x) or self.t_k != t:
                        
            g, h = self.cost.gradient(x, t, **kwargs), self.cost.hessian(x, t, **kwargs)
            d = self.cost.time_derivative(x, t, der="xt", **kwargs) \
              + self.cost.time_derivative(x, t, der="tx", **kwargs)
            
            # linear and constant terms of the prediction function
            b = g - np.dot(h, x) + 0.5*self.t_s*d
            
            if gradient_only:
                c = 0
            else:
                c = self.cost.function(x, t, **kwargs) - \
                    (0.5*self.t_s*d + g).T.dot(x) + 0.5*x.T.dot(h.dot(x)) + \
                    0.5*self.t_s**2*self.cost.time_derivative(x, t, der="tt", **kwargs) + \
                    self.t_s*self.cost.time_derivative(x, t, der="t", **kwargs)
            
            # store new prediction
            self.prediction = costs.Quadratic(h, b, c)
            self.x_k, self.t_k = x, t